import logging
import logging.config
from logging.handlers import RotatingFileHandler

LOGGER_NAME = 'dnfal'
LOGGER_FORMAT = '[%(asctime)s] %(levelname)s %(module)s:%(funcName)s:%(lineno)d "%(message)s"'

LEVEL_DEBUG = 'debug'
LEVEL_INFO = 'info'
LEVEL_WARNING = 'warning'
LEVEL_ERROR = 'error'
LEVEL_CRITICAL = 'critical'



logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'default': {
            'format': LOGGER_FORMAT
        }
    },
    'loggers': {
        LOGGER_NAME: {
            'level': 'INFO',
            'propagate': False,
            'handlers': [],
        },
    },
})

logger = logging.getLogger(LOGGER_NAME)


def config_logger(
    level: str,
    to_console: bool = True,
    file_path: str = ''
):
    logger.setLevel(level.upper())
    formatter = logging.Formatter(LOGGER_FORMAT)

    if to_console:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    if file_path:
        file_handler = RotatingFileHandler(
            file_path,
            maxBytes=100 * 1024,
            backupCount=10
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    if not to_console and not file_path:
        null_handler = logging.NullHandler()
        logger.addHandler(null_handler)



from cvtlib.video import VideoCapture

from .alignment import FaceAligner, FaceMarker
from .detection import FaceDetector
from .encoding import FaceEncoder
from dnfal.genderage import GenderAgePredictor
from .engine import FrameAnalyzer, VideoAnalyzer, FaceMatcher
from .loggers import logger, config_logger
from .settings import Settings


class FacesVision:
    def __init__(self, se: Settings):

        self.settings: Settings = se

        self.init_logger(se)

        # noinspection PyTypeChecker
        self._face_detector: FaceDetector = None
        # noinspection PyTypeChecker
        self._face_marker: FaceMarker = None
        # noinspection PyTypeChecker
        self._face_aligner: FaceAligner = None
        # noinspection PyTypeChecker
        self._face_encoder: FaceEncoder = None
        # noinspection PyTypeChecker
        self._face_matcher: FaceMatcher = None
        # noinspection PyTypeChecker
        self._genderage_predictor: GenderAgePredictor = None
        # noinspection PyTypeChecker
        self._frame_analyzer: FrameAnalyzer = None
        # noinspection PyTypeChecker
        self._video_capture: VideoCapture = None
        # noinspection PyTypeChecker
        self._video_analyzer: VideoAnalyzer = None

    @property
    def face_marker(self):
        if self._face_marker is None:
            self._face_marker = FaceMarker(
                weights_path=self.settings.marker_weights_path,
                force_cpu=self.settings.force_cpu
            )
            logger.info('Face marker created.')
        return self._face_marker

    @property
    def face_aligner(self):
        if self._face_aligner is None:
            se = self.settings
            self._face_aligner = FaceAligner(out_size=se.face_align_size)
            logger.info('Face aligner created.')
        return self._face_aligner

    @property
    def face_encoder(self):
        if self._face_encoder is None:
            self._face_encoder = FaceEncoder(
                weights_path=self.settings.encoder_weights_path,
                force_cpu=self.settings.force_cpu
            )
            logger.info('Face encoder created.')
        return self._face_encoder

    @property
    def face_detector(self):
        if self._face_detector is None:
            se = self.settings
            self._face_detector = FaceDetector(
                weights_path=se.detector_weights_path,
                min_height=se.detection_min_height,
                min_score=se.detection_min_score,
                nms_thresh=se.detection_nms_thresh,
                force_cpu=se.force_cpu
            )
            logger.info('Face detector created.')
        return self._face_detector

    @property
    def face_matcher(self):
        if self._face_matcher is None:
            self._face_matcher = FaceMatcher()
            logger.info('Face matcher created.')
        return self._face_matcher

    @property
    def genderage_predictor(self):
        if self._genderage_predictor is None:
            se = self.settings
            self._genderage_predictor = GenderAgePredictor(
                se.genderage_weights_path,
                force_cpu=se.force_cpu
            )
            logger.info('Gender-age predictor created.')
        return self._genderage_predictor

    @property
    def frame_analyzer(self):
        if self._frame_analyzer is None:
            se = self.settings

            face_marker = None
            face_aligner = None
            face_encoder = None

            if not se.detection_only or se.video_mode == VideoAnalyzer.MODE_HUNT:
                face_marker = self.face_marker
                face_aligner = self.face_aligner
                face_encoder = self.face_encoder

            self._frame_analyzer = FrameAnalyzer(
                detector=self.face_detector,
                marker=face_marker,
                aligner=face_aligner,
                encoder=face_encoder,
                detection_only=se.detection_only,
                max_frame_size=se.max_frame_size,
                store_frames=se.store_face_frames,
                max_deviation=se.align_max_deviation,
                marking_min_score=se.marking_min_score,
                face_padding=se.detection_face_padding
            )
            logger.info('Frame analyzer created.')

        return self._frame_analyzer

    @property
    def video_capture(self):
        if self._video_capture is None:
            se = self.settings
            self._video_capture = VideoCapture(
                src=se.video_capture_source,
                auto_grab=se.video_real_time
            )
            logger.info('Video capture created.')
        return self._video_capture

    @property
    def video_analyzer(self):
        if self._video_analyzer is None:
            se = self.settings
            self._video_analyzer = VideoAnalyzer(
                frame_analyzer=self.frame_analyzer,
                video_capture=self.video_capture,
                real_time=se.video_real_time,
                detect_interval=se.video_detect_interval,
                faces_memory=se.video_face_memory,
                similarity_thresh=se.similarity_thresh,
                mode=se.video_mode,
                hunt_embeddings=se.video_hunt_embeddings,
                hunt_keys=se.video_hunt_keys,
                start_at=se.video_start_at,
                stop_at=se.video_stop_at,
                roi_adapt=se.video_roi_adapt
            )
            logger.info('Video analyzer created.')

        return self._video_analyzer

    def update_settings(self, se: Settings):

        self.settings = se
        self.init_logger(se)

        # noinspection PyTypeChecker
        self._face_detector: FaceDetector = None
        # noinspection PyTypeChecker
        self._face_marker: FaceMarker = None
        # noinspection PyTypeChecker
        self._face_aligner: FaceAligner = None
        # noinspection PyTypeChecker
        self._face_encoder: FaceEncoder = None
        # noinspection PyTypeChecker
        self._face_matcher: FaceMatcher = None
        # noinspection PyTypeChecker
        self._genderage_predictor: GenderAgePredictor = None
        # noinspection PyTypeChecker
        self._frame_analyzer: FrameAnalyzer = None
        # noinspection PyTypeChecker
        self._video_capture: VideoCapture = None
        # noinspection PyTypeChecker
        self._video_analyzer: VideoAnalyzer = None

    @staticmethod
    def init_logger(se: Settings):
        if se.logging_file or se.log_to_console:
            config_logger(
                level=se.logging_level,
                file_path=se.logging_file,
                to_console=se.log_to_console
            )



from itertools import product as product
from math import ceil
from typing import Tuple

import numpy as np
import torch
from fnms import nms

from ._model import FaceBoxes

MIN_SIZES = [[32, 64, 128], [256], [512]]
STEPS = [32, 64, 128]
CLIP = False
VARIANCE = [0.1, 0.2]
KEEP_TOP_K = 5000
TRANSFORM_OFFSET = np.float32((-104, -117, -123))


class FaceDetector:
    """Face detector.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.

    force_cpu : bool, optional, (default=False)
        Whether to force model run on CPU even if CUDA is available.

    min_score : float, optional, (default=0.9)
        Minimum confidence score of detected faces. Faces with scores lower
        than the `min_score` will be discarded.

    nms_thresh : float, optional, (default=0.5)
        Non-maximum suppression threshold of detected faces. From the set of
        face boxes which overlaps more than the corresponding NMS threshold the
        one with the highest score will be selected and the others discarded.

    min_height : int, optional, (default=24)
        Minimum height in pixels of face to be detected. Faces with height
        lower than `min_height` will be discarded.
    """

    def __init__(
        self,
        weights_path: str,
        force_cpu: bool = False,
        min_score: float = 0.9,
        nms_thresh: float = 0.5,
        min_height: int = 24
    ):
        self.model: FaceBoxes = FaceBoxes(
            phase='test', size=None, num_classes=2
        )
        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)

        if self.gpu is None:
            cpu = torch.device('cpu')
            pretrained_dict = torch.load(weights_path, map_location=cpu)
        else:
            pretrained_dict = torch.load(weights_path)

        self.model.load_state_dict(pretrained_dict, strict=False)
        if self.gpu is not None:
            self.model = self.model.to(self.gpu)

        self.model.eval()
        # noinspection PyTypeChecker
        self.image_size: Tuple[int, int] = None
        # noinspection PyTypeChecker
        self.priors: torch.Tensor = None
        # noinspection PyTypeChecker
        self.scale: torch.Tensor = None

        self.min_score: float = min_score
        self.nms_thresh: float = nms_thresh
        self.min_height: int = min_height

    def detect(self, image: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Estimate face landmark locations.

        Parameters
        ----------
        image : array_like image
            Input image to be analyzed.

        Returns
        -------
        boxes : array_like of shape = [n_faces, 4] and `dtype=np.int32`
            Detected face boxes. Each entry is face box. The first two values
            in a face box represent the top-left corner coordinates, while the
            latest two values represents the bottom-left corner coordinates.
        scores : array_like of shape = [n_faces] and `dtype=np.float`
            The detection score for each detected face. The score is a number
            between 0 and 1 that represents the confidence of the detection.
        """

        h, w = image.shape[0:2]
        face_boxes = np.array([])
        face_scores = np.array([])

        if self.image_size != (w, h):
            self._size_updated((w, h))
            self.image_size = (w, h)

        image = self._image_transform(image)

        locations, confidence = self.model(image)
        boxes = _decode(locations.data.squeeze(0), self.priors.data, VARIANCE)
        boxes *= self.scale
        boxes = boxes.cpu().numpy()
        scores = confidence.squeeze(0).data.cpu().numpy()[:, 1]

        if len(boxes) == 0:
            return face_boxes, face_scores

        # ignore low scores
        inds = np.where(scores > self.min_score)[0]

        if len(inds) == 0:
            return face_boxes, face_scores

        boxes = boxes[inds]
        scores = scores[inds]

        # keep top-K before NMS
        order = scores.argsort()[::-1][:KEEP_TOP_K]
        boxes = boxes[order]
        scores = scores[order]

        # do NMS
        dets = np.hstack(
            (boxes, scores[:, np.newaxis])
        ).astype(np.float32, copy=False)
        keep = nms(dets, self.nms_thresh, force_cpu=(self.gpu is None))
        dets = dets[keep, :]

        # keep only boxes with height higher than min_height
        heights = dets[:, 3] - dets[:, 1]
        keep = np.where(heights >= self.min_height)[0]
        dets = dets[keep, :]

        boxes = dets[:, 0:4].astype(np.int32)
        np.clip(boxes, a_min=(0, 0, 0, 0), a_max=(w, h, w, h), out=boxes)
        scores = dets[:, 4]

        return boxes, scores

    def _image_transform(self, image: np.ndarray):
        if image.dtype != np.float32:
            image = image.astype(np.float32)
        image += TRANSFORM_OFFSET
        image = image.transpose((2, 0, 1))
        image = torch.from_numpy(image).unsqueeze(0)
        if self.gpu is not None:
            image = image.to(self.gpu)
        return image

    def _size_updated(self, size):
        priors = _prior_boxes((size[1], size[0]))
        scale = torch.tensor([size[0], size[1], size[0], size[1]])
        if self.gpu is not None:
            priors = priors.to(self.gpu)
            scale = scale.to(self.gpu)
        self.priors = priors
        self.scale = scale


def _prior_boxes(shape: Tuple[int, int]) -> torch.Tensor:
    feature_maps = [
        [ceil(shape[0] / step), ceil(shape[1] / step)]
        for step in STEPS
    ]
    anchors = []
    for k, f in enumerate(feature_maps):
        min_sizes = MIN_SIZES[k]
        for i, j in product(range(f[0]), range(f[1])):
            for min_size in min_sizes:
                s_kx = min_size / shape[1]
                s_ky = min_size / shape[0]
                if min_size == 32:
                    dense_cx = [
                        x * STEPS[k] / shape[1]
                        for x in [j + 0, j + 0.25, j + 0.5, j + 0.75]
                    ]
                    dense_cy = [
                        y * STEPS[k] / shape[0]
                        for y in [i + 0, i + 0.25, i + 0.5, i + 0.75]
                    ]
                    for cy, cx in product(dense_cy, dense_cx):
                        anchors += [cx, cy, s_kx, s_ky]
                elif min_size == 64:
                    dense_cx = [
                        x * STEPS[k] / shape[1]
                        for x in [j + 0, j + 0.5]
                    ]
                    dense_cy = [
                        y * STEPS[k] / shape[0]
                        for y in [i + 0, i + 0.5]
                    ]
                    for cy, cx in product(dense_cy, dense_cx):
                        anchors += [cx, cy, s_kx, s_ky]
                else:
                    cx = (j + 0.5) * STEPS[k] / shape[1]
                    cy = (i + 0.5) * STEPS[k] / shape[0]
                    anchors += [cx, cy, s_kx, s_ky]

    output = torch.tensor(anchors).view(-1, 4)
    if CLIP:
        output.clamp_(max=1, min=0)
    return output


def _decode(loc, priors, variances):
    """Decode locations from predictions using priors to undo
    the encoding we did for offset regression at train time.
    Args:
        loc (tensor): location predictions for loc layers,
            Shape: [num_priors,4]
        priors (tensor): Prior boxes in center-offset form.
            Shape: [num_priors,4].
        variances: (list[float]) Variances of priorboxes
    Return:
        decoded bounding box predictions
    """

    boxes = torch.cat((
        priors[:, :2] + loc[:, :2] * variances[0] * priors[:, 2:],
        priors[:, 2:] * torch.exp(loc[:, 2:] * variances[1])), 1)
    boxes[:, :2] -= boxes[:, 2:] / 2
    boxes[:, 2:] += boxes[:, :2]
    return boxes



from typing import List

import cv2 as cv
import numpy as np
import torch
import torch.nn as nn

from ._model import DenseNet

DENSE_NET_CLS_NUM = 751

SCALE = np.float32([0.01712475, 0.017507, 0.01742919])
OFFSET = np.float32([2.11790393, 2.03571429, 1.80444444])
INPUT_WIDTH = 256
INPUT_HEIGHT = 128


class PersonEncoder:
    """Full-body person encoder for re-identification.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.
    """

    def __init__(
        self,
        weights_path: str,
        force_cpu: bool = False
    ):
        self.model: DenseNet = DenseNet(class_num=DENSE_NET_CLS_NUM)

        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)

        if self.gpu is None:
            cpu = torch.device('cpu')
            pretrained_dict = torch.load(weights_path, map_location=cpu)
        else:
            pretrained_dict = torch.load(weights_path)

        self.model.load_state_dict(pretrained_dict)

        self.model.classifier.classifier = nn.Sequential()

        if self.gpu is not None:
            self.model = self.model.to(self.gpu)

        self.model.eval()

    def encode(self, images: List[np.ndarray]):
        """Encode face images.

        Parameters
        ----------
        images : list of array_like images of length = n_persons
            A list of face images to be encoded.

        Returns
        -------
        embeddings : array_like of shape = [n_persons, 512]
            Embeddings vectors. Each entry is the embedding vector of the
            corresponding input person image.
        """

        batch_shape = (2 * len(images), 3, INPUT_HEIGHT, INPUT_WIDTH)
        batch = np.zeros(batch_shape, dtype=np.float32)

        for index, image in enumerate(images):
            batch[2 * index, :, :, :] = _image_transform(image)
            batch[2 * index + 1, :, :, :] = _image_transform(cv.flip(image, 1))

        batch_tensor = torch.from_numpy(batch)
        if self.gpu is not None:
            batch_tensor = batch_tensor.to(self.gpu)

        with torch.no_grad():
            embeddings = self.model(batch_tensor)
            embeddings = embeddings[0::2] + embeddings[1::2]
            embeddings_norm = torch.norm(embeddings, p=2, dim=1, keepdim=True)
            embeddings = embeddings.div(embeddings_norm.expand_as(embeddings))

        if self.gpu is not None:
            embeddings = embeddings.cpu()

        return embeddings.numpy()


def _image_transform(image: np.ndarray):

    # BGR to RGB
    image = cv.resize(image, (INPUT_WIDTH, INPUT_HEIGHT))
    image = image[:, :, ::-1]
    if image.dtype != np.float32:
        image = np.float32(image)
    image = SCALE * image - OFFSET
    image = image.transpose((2, 0, 1))
    image = np.reshape(image, (1, 3, INPUT_HEIGHT, INPUT_WIDTH))

    return image





from time import time, sleep
from typing import Dict
from typing import List, Tuple

import cv2 as cv
import cvtlib
import numpy as np
import sklearn
from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import DBSCAN, OPTICS
from scipy.sparse.csgraph import connected_components
from sklearn.cluster import AgglomerativeClustering
import torch
from cvtlib.video import VideoCapture

from .mtypes import Face, Frame, Subject
from .alignment import FaceAligner, FaceMarker
from .detection import FaceDetector
from .encoding import FaceEncoder
from .loggers import logger

EMBEDDINGS_LENGTH = 512


def distance_to_similarity(dists: (np.ndarray, float)):
    a = 5
    b = 1.3
    return 1 / (1 + np.exp(a * (dists - b)))


def similarity_to_distance(sim: (np.ndarray, float)):
    a = 5
    b = 1.3
    sim = np.clip(sim, a_max=(1 - 1e-4), a_min=1e-4)
    return np.clip((1/a) * np.log(1/sim - 1) + b, a_min=0, a_max=None)


class AdaptiveRoi:
    """Adaptive Region-of-Interest (ROI) generator.

    Iteratively computes a ROI by finding a rectangle that encloses a set of
    other rectangles successively added at each iteration. Initially the ROI
    is empty. ROI is updated by adding a new rectangles. When a new
    rectangle is added, the ROI is updated to a new rectangle that contains
    both, the actual ROI rectangle an the new rectangle added. At each step
    the moving average deviation of the ROI is computed. The smoothness of
    this average is controlled via a smoothing factor `alpha`. The ROI is
    considered valid only when the moving average deviation is below a
    threshold.

    An adaptive ROI can help to save considerable processing time in video
    object detection tasks, when the video comes from a static camera. In
    these cases the location of the detected objects in the video frames are
    likely to be located inside a rectangle that changes little (in size and
    location) over time. So, instead analyzing the entire frame, only a
    sub-region of the frame is analyzed.

    Parameters
    ----------
    thresh : float, optional, (default=1)
        Convergence threshold for the moving average of the ROI deviation.

    alpha : float, optional, (default=0.1)
        Smoothing factor of the exponential moving average of the ROI
        deviation.
    """
    def __init__(
        self,
        thresh: float = 1,
        alpha: float = 0.1
    ):
        self.alpha: float = alpha
        self.thresh: float = thresh

        self._cur_roi: [tuple, None] = None
        self._avg_roi: [tuple, None] = None
        self._roi_dev: float = float('inf')

    @property
    def roi(self):
        """Current ROI.

        The current ROI is a tuple of four float numbers representing a
        rectangle that encloses all the previously added rectangles, if the
        exponential moving average of its deviation have converged. Else it
        is `None`. The first two values in the tuple rectangle represents
        the top-left coordinates of the rectangle, while the latest two
        values represents the bottom-left coordinates.

        Returns
        -------
        curr_roi : tuple of length=4 or None
        """
        return self._cur_roi if self._roi_dev <= self.thresh else None

    def add_box(self, box: tuple):
        """Adds a new rectangle.

        Parameters
        -------
        box : tuple of length=4
            The first two values in the tuple rectangle represents the top-left
            coordinates of the rectangle, while the latest two values
            represents the bottom-left coordinates.
        """
        cur_roi = self._cur_roi
        avg_roi = self._avg_roi
        alpha = self.alpha

        if cur_roi is None:
            cur_roi = tuple(box)
        else:
            cur_roi = (
                min(cur_roi[0], box[0]),
                min(cur_roi[1], box[1]),
                max(cur_roi[2], box[2]),
                max(cur_roi[3], box[3]),
            )

        self._cur_roi = cur_roi

        if avg_roi is None:
            self._avg_roi = tuple(cur_roi)
            self._roi_dev = sum(cur_roi) / 4
        else:
            self._avg_roi = tuple([
                alpha * cur_roi[i] + (1 - alpha) * avg_roi[i]
                for i in range(4)
            ])

            roi_dev = sum((
                max(avg_roi[0] - cur_roi[0], 0),
                max(avg_roi[1] - cur_roi[1], 0),
                max(cur_roi[2] - avg_roi[2], 0),
                max(cur_roi[3] - avg_roi[3], 0)
            )) / 4

            self._roi_dev = alpha * roi_dev + (1 - alpha) * self._roi_dev


class FaceMatcher:
    """Face matcher.

    Parameters
    ----------
    similarity_thresh : float, optional, (default=0.5)
        Similarity threshold for face matching. Two faces are considered to
        belong to the same person if the similarity between them is higher
        than `similarity_thresh`. A similarity of `1` means that the two
        faces are exactly the same, while a similarity of `0` means that the
        two faces are very different. Similarity between two faces if
        computed on the basis of the distance between their embeddings
        vectors.
    """
    def __init__(
        self,
        similarity_thresh: float = 0.5
    ):
        self._matcher = cv.BFMatcher(cv.NORM_L2)
        self._distance_thresh: float
        self.similarity_thresh = similarity_thresh

    def match(self, x_test, x_train, y_train):
        """Match two sets of face embeddings vectors

        Parameters
        ----------
        x_test : array-like of shape = [n_tests, n_features]
            The test set of face embeddings vectors. Each entry represents a
            embeddings vector of a face.

        x_train : array-like of shape = [n_trains, n_features]
            The train set of face embeddings vectors. Each entry represents a
            embeddings vector of a face.

        y_train : array-like of shape = [n_trains]
            The train set of face labels. Each entry represents the label of
            the corresponding entry in `x_train`.

        Returns
        -------
        y_test : list of length = n_trains
            The matched face labels. Each entry represents the matched label
            of the corresponding entry in `x_test`.

        similarity : list of length = n_trains
            The similarity scores of matched face labels. Each entry
            represents the similarity score of the the corresponding entry
            in `x_test`.
        """

        matches = self._matcher.radiusMatch(
            x_test, x_train, self._distance_thresh
        )

        y_test = []
        similarity = []
        for i, matches_i in enumerate(matches):
            y = [y_train[m.trainIdx] for m in matches_i]
            d = np.float32([m.distance for m in matches_i])

            y, ind, count = np.unique(y, return_index=True, return_counts=True)
            d = d[ind] / count

            if len(y):
                ind = np.argsort(d)
                y = y[ind]
                d = d[ind]

            y_test.append(y)
            similarity.append(distance_to_similarity(d))

        return y_test, similarity

    @property
    def similarity_thresh(self):
        return distance_to_similarity(self._distance_thresh)

    @similarity_thresh.setter
    def similarity_thresh(self, value: float):
        self._distance_thresh = similarity_to_distance(value)


class FrameAnalyzer:
    """Frame analyzer.

    Parameters
    ----------
    detector : FaceDetector
        A FaceDetector object.

    marker : FaceMarker or None
        A FaceMarker object.

    aligner: FaceAligner or None
        A FaceAligner object.

    encoder: FaceEncoder or None
        FaceEncoder object.

    detection_only: bool, optional, (default=False)
        Face analysis includes three steps: detection, alignment and
        encoding. If `detection_only = True`, only the detection step is
        done and `marker`, `aligner` and `encoder` can be set all to None.

    max_frame_size: int, optional, (default=0)
        If `store_frame = True` then the input image will be stored in each
        Face object created from a detected face. If the input image is
        stored, it will be resized so that is maximum dimension is lowest
        than `max_frame_size`. Set it to a value lower than or equal to zero
        to do not resize.

    store_frames: bool, optional, (default=False)
        If `store_frame = True` then the input image will be stored in each
        Face object created from a detected face.

    max_deviation: tuple of length 2 or None, optional, (default=None)
        Face analysis includes three steps: detection, alignment and
        encoding. At the alignment step, the nose deviation is computed. The
        nose deviation is a tuple where the first and second elements
        represents the horizontal and vertical deviation of nose landmark,
        respectively. If this deviation is higher than `align_max_deviation`
        (element-wise comparision), the face is discarded. For a deviation
        lower than ( 0.4, 0.3), the face is approximately frontal in both
        planes. Set it to `None` to allow any deviation.

    marking_min_score: float, optional, (default=0.9)
        Face analysis includes three steps: detection, alignment and
        encoding. At the alignment step, face landmarks are computed and a
        marking score between 0 and 1 is assigned to each face. Lower scores
        mean lower face alignment confidence. Faces with marking score lower
        than the `marking_min_score` will be discarded.

    face_padding: float, optional, (default=0.2)
        After a face is detected, the corresponding face image is extracted
        from the original image and sent to a face alignment stage. Padding the
        detected face box can help to obtain better alignment results. Padding
        value is a number between 0 and 1, relative to the face box height.
    """

    def __init__(
        self,
        detector: FaceDetector,
        marker: [FaceMarker, None],
        aligner: [FaceAligner, None],
        encoder: [FaceEncoder, None],
        detection_only: bool = False,
        max_frame_size: int = 0,
        store_frames: bool = False,
        store_aligned: bool = False,
        max_deviation: Tuple[float, float] = None,
        marking_min_score: float = 0.9,
        face_padding: float = 0.2
    ):
        self.face_detector: FaceDetector = detector
        self.face_marker: FaceMarker = marker
        self.face_aligner: FaceAligner = aligner
        self.face_encoder: FaceEncoder = encoder
        self.detection_only: bool = detection_only
        self.max_frame_size: int = max_frame_size
        self.store_frames: bool = store_frames
        self.store_aligned: bool = store_aligned
        self.max_deviation: Tuple[float, float] = max_deviation
        self.marking_min_score: float = marking_min_score
        self.face_padding: float = face_padding

    def find_faces(
        self,
        image: np.ndarray,
        timestamp: float = 0
    ):
        """Detects, align and encode faces in a image.

        Parameters
        ----------
        image : array-like of `dtype=np.uint8`
            Input image to be analyzed.

        timestamp : float, optional, (default=0)
            A timestamp to be stored in each of the created Face objects.

        Returns
        -------
        faces : list of mtypes.Face objects of length=n_faces
            A list of a Face objects, each of them containing information
            about a detected face in the input image.
        embeddings : array_like of shape = [n_faces, 512]
            Embeddings vectors of detected faces. Each entry is the  embedding
            vector of the corresponding face in `faces`.
        """

        h, w = image.shape[0:2]
        faces: List[Face] = []
        embeddings: np.ndarray = np.array([])

        boxes, detect_scores = self.face_detector.detect(image)
        n_boxes = len(boxes)

        if n_boxes == 0:
            return faces, embeddings

        face_images = [image[b[1]:b[3], b[0]:b[2]] for b in boxes]

        padded_face_images = face_images
        padded_boxes = boxes
        if self.face_padding != 0:
            pad = self.face_padding
            paddings = [int(pad * (b[2] - b[0])) for b in boxes]
            padded_boxes = [(
                max(0, b[0] - p),
                max(0, b[1] - p),
                min(b[2] + p, w - 1),
                min(b[3] + p, h - 1),
            ) for b, p in zip(boxes, paddings)]
            offsets = [
                (box[0] - padded_box[0], box[1] - padded_box[1])
                for box, padded_box in zip(boxes, padded_boxes)
            ]
            padded_face_images = [
                image[b[1]:b[3], b[0]:b[2]]
                for b in padded_boxes
            ]
        else:
            offsets = [(0, 0)] * n_boxes

        padded_boxes = [
            (b[0] / w, b[1] / h, b[2] / w, b[3] / h)
            for b in padded_boxes
        ]

        frame = None
        if self.store_frames and n_boxes:
            frame_image = image
            if max(image.shape[0:2]) > self.max_frame_size > 0:
                frame_image, _ = cvtlib.image.resize(
                    image, self.max_frame_size
                )
            frame = Frame(image=frame_image)

        if self.detection_only:
            for i in range(n_boxes):
                faces.append(Face(
                    image=padded_face_images[i],
                    box=padded_boxes[i],
                    frame=frame,
                    detect_score=detect_scores[i],
                    timestamp=timestamp,
                    offset=offsets[i]
                ))
        else:
            face_marks, mark_scores = self.face_marker.mark(face_images)
            aligned_face_images = []
            max_deviation = self.max_deviation
            marking_min_score = self.marking_min_score
            for i in range(n_boxes):
                face_mark = face_marks[i] + offsets[i]
                aligned_face_image, nose_deviation = self.face_aligner.align(
                    padded_face_images[i], face_mark
                )
                if (mark_scores[i] > marking_min_score) and (
                    max_deviation is None or (
                        nose_deviation[0] <= max_deviation[0]) and (
                        nose_deviation[1] <= max_deviation[1]
                    )
                ):
                    aligned_face_images.append(aligned_face_image)

                    face = Face(
                        image=padded_face_images[i],
                        box=padded_boxes[i],
                        frame=frame,
                        landmarks=face_mark,
                        nose_deviation=nose_deviation,
                        detect_score=detect_scores[i],
                        mark_score=mark_scores[i],
                        timestamp=timestamp,
                        offset=offsets[i]
                    )

                    if self.store_aligned:
                        face.aligned_image = aligned_face_image

                    faces.append(face)

            if len(faces) and self.face_encoder is not None:
                embeddings = self.face_encoder.encode(aligned_face_images)
                for i, face in enumerate(faces):
                    face.embeddings = embeddings[i]

        return faces, embeddings


VIDEO_PAUSE_SLEEP = 3
VIDEO_DEFAULT_FRAME_RATE = 24


class VideoAnalyzer:
    """Video analyzer.

    Parameters
    ----------

    frame_analyzer: FrameAnalyzer
        A FrameAnalyzer object.

    video_capture: cvtlib.VideoCapture
        A VideoCapture object.

    real_time: bool, optional, (default=True)
        Whether the video capture is from a camera or not. Video analysis can
        be done from a real time source, like a camera, or from a non-real time
        source, like a video file. Although the video analysis for both types
        of sources is similar,some differences exists. For example, for a real
        time source when the algorithm finish analyzing one frame, it waits
        (sleeps) a time equal to `detect_interval` before processing a new
        frame. For a video file, the algorithms jumps directly into the next
        target frame after analyzing the current frame.

    detect_interval: float, optional (default=1)
        Interval in seconds to analyze frames in video processing. In order to
        save processing time, not all frames are analyzed during video
        processing. For a real time source, when the algorithm finish analyzing
        one frame, it waits (sleeps) a time equal to `detect_interval` before
        processing a new frame. Similarly, for a video file, the algorithms
        jumps directly into the next target frame located
        `video_detect_interval` seconds beyond. Setting it to a reasonable
        value can help to save considerable processing time, taking into
        account the (typical) low movement speed of people.

    faces_memory: float, optional (default=30)
        Time memory in seconds of faces during video analysis. During video
        analysis, the algorithm try to match detected faces and group matched
        faces into Subject objects. A face is added to a Subject object only if
        no face have been added to that subject in the previous `face_memory`
        seconds. This setting helps to control the growth without limits of the
        number of faces added to a subject. Also, as face images belonging to
        the same person in successive nearby frames may be very similar, it
        also avoid storing redundant information.

    similarity_thresh: float, optional (default=0.5)
        Similarity threshold for face matching. During face matching, two faces
        are considered to belong to the same person if the similarity between
        them is higher than `similarity_thresh`. A similarity of `1` means that
        the two faces are exactly the same, while a similarity of `0` means
        that the two faces are very different. Similarity between two faces if
        computed on the basis of the distance between their embeddings vectors.

    mode: str, optional, (default='all')
        Video analysis mode. Video analysis can be executed in two different
        modes: "all" or "hunt". In the "all" mode, the algorithm tries to
        register all the detected faces, grouping them in Subject objects. In
        the "hunt" mode the algorithm tries to match the detected faces with a
        predefined list of face embeddings and to register only matched faces.

    hunt_embeddings: array_like of shape = [n_hunts, 512] or None
        Embeddings vectors to hunt for. If `mode = "hunt"` the algorithm tries
        to match the detected faces against a set of embeddings vectors in
        `hunt_embeddings` that represents the hunted faces, and to register
        only matched faces. Each embedding vector is of length 512, as returned
        by encoding.FaceEncoder.

    hunt_keys: array_like of shape = [n_hunts] or None
        List of face keys associated to the hunt embeddings. If `video_mode =
        "hunt"` the algorithm tries to match the detected faces against a set
        of embeddings vectors in `hunt_embeddings` that represents the hunted
        faces, and to register only matched faces. Each embedding vector of a
        hunted face must be associated to face key in `hunt_keys`, which
        is a list of integers numbers or strings of length `n_hunts`.

    start_at: float, optional, (default=0)
        Starting time in seconds of processing for video source files. If
        `real_time = False`, the video file will be analyzed starting from
        `start_at` seconds. Note that if the duration of the video file is
        lower than `start_at` or `start_at` is higher than `stop_at`,
        `video_start_at` will be ignored.

    stop_at: float, optional, (default=-1)
        Stopping time in seconds of processing for video source files. If
        `real_time = False`, the video file will analysis will be terminated at
        `stop_at` seconds. Note that if the duration of the video file is lower
        than `stop_at` or `start_at` is higher than `stop_at`, `stop_at` will
        be ignored. Set it to a negative value to run the analysis until the
        end of the video file.

    roi_adapt: bool, optional, (default=False)
        Use adaptive ROI during video analysis. If set to `True`, adaptive
        Region-Of-Interest (ROI) will be used during video analysis. This can
        help to save considerable processing time if the video comes from a
        static camera, as in these case the faces in the video frames are
        likely to be located inside a rectangle that changes little (in size
        and  location) over time. So, instead analyzing the entire frame, only
        a sub-region of the frame is analyzed. This rectangle is estimated at
        run time an used once it converges.
    """

    LOGGING_FRAME_INTERVAL = 100
    MAX_CAMERA_NULL_FRAMES = 50
    MAX_VIDEO_NULL_FRAMES = 3
    MAX_STORED_SUBJECTS = 20
    STORED_SUBJECTS_SHRINK = 0.3

    MODE_HUNT = 'hunt'
    MODE_ALL = 'all'

    def __init__(
        self,
        frame_analyzer: FrameAnalyzer,
        video_capture: VideoCapture,
        real_time: bool = True,
        detect_interval: float = 1,
        faces_memory: float = 30,
        similarity_thresh: float = 0.5,
        mode: str = MODE_ALL,
        hunt_embeddings: np.ndarray = None,
        hunt_keys: list = None,
        start_at: float = 0,
        stop_at: float = -1,
        roi_adapt: bool = False
    ):

        self.frame_analyzer: FrameAnalyzer = frame_analyzer
        self.video_capture: VideoCapture = video_capture
        self.video_capture.open()
        self.real_time: bool = real_time
        self.faces_memory: float = faces_memory
        self.similarity_thresh: float = similarity_thresh
        self.detect_interval: float = detect_interval
        self.mode: str = mode
        self.start_at: float = start_at
        self.stop_at: float = stop_at
        self.roi_adapt: bool = roi_adapt

        self.hunt_keys = hunt_keys

        if self.mode == self.MODE_HUNT:
            error = ValueError(
                f'Invalid "hunt_embeddings" parameter. It must be a list of lists of numbers, '
                f'each one having {EMBEDDINGS_LENGTH} elements.'
            )
            try:
                self.hunt_embeddings: np.ndarray = np.array(hunt_embeddings, np.float32)
            except ValueError:
                raise error

            if len(self.hunt_embeddings) == 0:
                logger.warning('Mode is set to "hunt" and "hunt_embeddings" parameter is zero length.')

            if len(self.hunt_embeddings) and (
                len(self.hunt_embeddings.shape) != 2 or
                self.hunt_embeddings.shape[1] != EMBEDDINGS_LENGTH
            ):
                raise error

            if self.hunt_keys is not None:
                if len(self.hunt_keys) != len(self.hunt_embeddings):
                    raise ValueError('Parameters "hunt_keys" and "hunt_embeddings" have different lengths.')

        self._run: bool = False
        self._paused: bool = False
        self.timestamp: float = 0
        self.frame: np.ndarray = np.array([])
        self.stated_at: float = -1

        self.subjects: Dict[int, Subject] = {}
        self.embeddings: List[np.ndarray] = []
        self.keys: List[int] = []
        self.key_count: int = 0

        if not self.real_time:
            frame_rate = VIDEO_DEFAULT_FRAME_RATE
            try:
                frame_rate = self.video_capture.frame_rate
            except AttributeError:
                logger.warning(
                    f'Video frame rate could not be determined. '
                    f'Assuming {VIDEO_DEFAULT_FRAME_RATE} fps by default.'
                )
            self.frame_interval = int(frame_rate * self.detect_interval)

            self._check_time_control()

        self.frame_rate: float = 0
        self.frames_count: int = 0
        self.faces_count: int = 0
        self.processing_time: float = 0

        self.matcher: cv.BFMatcher = cv.BFMatcher(cv.NORM_L2)
        self.dist_thresh = similarity_to_distance(self.similarity_thresh)

        self.adaptive_roi = AdaptiveRoi(thresh=0.025)

    def stop(self):
        """Stop video analysis"""

        self._run = False
        self._paused = False

    def pause(self):
        """Pause video analysis"""

        self._paused = True

    def resume(self):
        """Resume video analysis"""

        self._paused = False

    def run(
        self,
        frame_callback: callable = None,
        update_subject_callback: callable = None
    ):
        """Run video analysis.

        Parameters
        ----------
        frame_callback : callable or None, optional, (default=None)
            A function called every time a video frame is analyzed. It

        update_subject_callback : callable or None, optional, (default=None)
            A function called every time a video frame is analyzed. It
        """

        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        self._run = True
        self._paused = False
        self.key_count = 0

        last_detect_at = 0
        self.frames_count = 0
        self.processing_time = 0

        cap = self.video_capture

        if self.start_at > 0 and not self.real_time:
            try:
                cap.goto_time(self.start_at)
            except AttributeError:
                self.start_at = -1
                logger.warning(
                    'Video can be advanced, ignoring parameter "start_at".'
                )
            try:
                _ = cap.timestamp
            except AttributeError:
                self.stop_at = -1
                logger.warning(
                    'Video can be advanced, ignoring parameter "stop_at".'
                )

        logger.info('Starting video analysis.')

        if self.mode == self.MODE_HUNT:
            for index, embedding in enumerate(self.hunt_embeddings):
                self.subjects[index] = Subject(
                    faces=[],
                    embeddings=embedding
                )
                if self.hunt_keys is not None:
                    self.subjects[index].data['hunt_key'] = self.hunt_keys[index]

        cap_fail_counter = 0
        cap_success = False
        do_detection = False
        self.start_at = time()

        while self._run:

            if self.real_time:
                wait = last_detect_at + self.detect_interval - time()
                if wait > 0:
                    sleep(wait)
                do_detection = True

            while self._paused:
                sleep(VIDEO_PAUSE_SLEEP)

            self.frame = None

            if self.real_time:
                self.timestamp = time()
            else:
                try:
                    self.timestamp = cap.timestamp
                except AttributeError:
                    frame_rate = VIDEO_DEFAULT_FRAME_RATE
                    try:
                        frame_rate = cap.frame_rate
                    except AttributeError:
                        pass
                    self.timestamp = cap.frame_number / frame_rate

            if not self.real_time and (0 < self.stop_at < self.timestamp):
                logger.info('Stopping analysis, "stop_at" time reached.')
                break

            if self.real_time:
                # do_detection = (
                #     self.timestamp - last_detect_at >= self.detect_interval
                # )
                if do_detection:
                    self.frame, cap_success = cap.next_frame()
            else:
                do_detection = cap.frame_number % self.frame_interval == 0
                if do_detection:
                    self.frame, cap_success = cap.next_frame()
                else:
                    cap_success = cap.grab_next()
            #do_detection = False
            if do_detection and self.frame is not None:
                cap_fail_counter = 0
                last_detect_at = self.timestamp

                frame_h, frame_w, = self.frame.shape[0:2]

                roi_box = None
                if self.roi_adapt:
                    roi_box = self.adaptive_roi.roi

                roi_frame = self.frame
                if roi_box is not None:
                    roi_box = (
                        int(frame_w * roi_box[0]),
                        int(frame_h * roi_box[1]),
                        int(frame_w * roi_box[2]),
                        int(frame_h * roi_box[3]),
                    )
                    roi_frame = self.frame[
                        roi_box[1]:roi_box[3],
                        roi_box[0]:roi_box[2]
                    ]

                faces, embeddings = self.frame_analyzer.find_faces(
                    roi_frame, timestamp=self.timestamp
                )

                if self.roi_adapt:
                    roi_h, roi_w = roi_frame.shape[0:2]
                    for face in faces:
                        if roi_box is not None:
                            face.box = (
                                (roi_box[0] + int(roi_w * face.box[0]))/frame_w,
                                (roi_box[1] + int(roi_h * face.box[1]))/frame_h,
                                (roi_box[0] + int(roi_w * face.box[2]))/frame_w,
                                (roi_box[1] + int(roi_h * face.box[3]))/frame_h,
                            )
                        self.adaptive_roi.add_box(face.box)

                if len(faces) > 0:
                    self.faces_count += len(faces)

                    # if self.timestamp < 4:
                    #     for i, face in enumerate(faces):
                    #         cv.imwrite(f'/home/ronin/Projects/active/dnfal/demos/output/face_hunt/face_{self.timestamp}_{i}.jpg', face.image)

                    if self.frame_analyzer.detection_only:
                        self._create_subjects(faces, update_subject_callback)
                    else:
                        if self.mode == self.MODE_ALL:
                            self._step_mode_all(
                                faces, embeddings, update_subject_callback
                            )
                        elif self.mode == self.MODE_HUNT:
                            self._step_mode_hunt(
                                faces, embeddings, update_subject_callback
                            )

                self.processing_time = (time() - self.start_at)
                self.frames_count += 1
                self.frame_rate = self.frames_count / self.processing_time
                logger.debug(f'Current frame rate: {self.frame_rate:.1f} FPS.')

            if not cap_success:
                cap_fail_counter += 1
                if (
                    not self.real_time and
                    cap_fail_counter >= self.MAX_VIDEO_NULL_FRAMES
                ):
                    break
                elif (
                    self.real_time and
                    cap_fail_counter >= self.MAX_CAMERA_NULL_FRAMES
                ):
                    raise RuntimeError(
                        'Maximum number of readed null frames reached.'
                    )

                sleep(1)

            if self.frame is not None and frame_callback is not None:
                frame_callback()
        # while self._run:

        logger.info('Video analysis finished.')

        if torch.cuda.is_available():
            torch.cuda.empty_cache()

    def _step_mode_all(
        self,
        faces: List[Face],
        embeddings: np.ndarray,
        update_subject_callback: callable
    ):
        if len(self.subjects) == 0:
            self._create_subjects(faces, update_subject_callback)
            self.embeddings.extend(list(embeddings))
            self.keys.extend(self.subjects.keys())
        else:
            matches = self.matcher.knnMatch(
                embeddings,
                np.array(self.embeddings),
                1
            )
            update_faces = []
            update_keys = []
            create_faces = []
            index_create = []

            for match in matches:
                if len(match):
                    if match[0].distance < self.dist_thresh:
                        update_faces.append(faces[match[0].queryIdx])
                        update_keys.append(self.keys[match[0].trainIdx])
                    else:
                        create_faces.append(faces[match[0].queryIdx])
                        index_create.append(match[0].queryIdx)

            if len(update_faces):
                self._update_subjects(
                    update_keys,
                    update_faces,
                    update_subject_callback
                )

            if len(create_faces):
                created_keys = self._create_subjects(
                    create_faces,
                    update_subject_callback
                )
                self.embeddings.extend(list(embeddings[index_create]))
                self.keys.extend(created_keys)
        self._check_subjects_size()

    def _step_mode_hunt(
        self,
        faces: List[Face],
        embeddings: np.ndarray,
        update_subject_callback: callable
    ):
        matches = self.matcher.knnMatch(embeddings, self.hunt_embeddings, 1)
        update_faces = []
        update_keys = []
        for i, match in enumerate(matches):
            if match[0].distance < self.dist_thresh:
                update_faces.append(faces[match[0].queryIdx])
                update_keys.append(match[0].trainIdx)

        self._update_subjects(
            update_keys,
            update_faces,
            update_subject_callback
        )

    def _create_subjects(self, faces, update_subject_callback=None):
        created_keys = []
        store = not self.frame_analyzer.detection_only
        for face in faces:
            self.key_count += 1
            key = self.key_count
            subject = Subject(
                faces=[face],
                embeddings=face.embeddings,
                key=key
            )
            face.subject = subject
            if update_subject_callback is not None:
                update_subject_callback(face)
            if store:
                self.subjects[key] = subject
                created_keys.append(key)

        return created_keys

    def _update_subjects(self, keys, faces, update_subject_callback=None):
        for key, face in zip(keys, faces):
            subject = self.subjects[key]
            if (self.timestamp - subject.last_updated) > self.faces_memory:
                self.subjects[key].append_face(face)
                face.subject = self.subjects[key]
                if update_subject_callback is not None:
                    update_subject_callback(face)

    def _check_subjects_size(self):
        total = len(self.subjects)
        del_count = int(self.STORED_SUBJECTS_SHRINK * total)
        if del_count and total > self.MAX_STORED_SUBJECTS:
            delete_keys = self.keys[0:del_count]
            for key in delete_keys:
                del self.subjects[key]

            del self.keys[0:del_count]
            del self.embeddings[0:del_count]

    def _check_time_control(self):
        if self.start_at > 0 or self.stop_at > 0:
            if (
                self.start_at > 0 and self.stop_at > 0
            ) and self.stop_at < self.start_at:
                logger.warning(
                    'Parameter "stop_at" is lower than "start_at", ignoring.'
                )
                self.start_at = 0
                self.stop_at = -1
            else:
                try:
                    video_duration = self.video_capture.duration_seconds
                    if self.start_at > 0 and self.start_at > video_duration:
                        logger.warning(
                            'Parameter "start_at" is higher than video '
                            'duration, ignoring it.'
                        )
                        self.start_at = 0
                    if self.stop_at > 0 and self.stop_at > video_duration:
                        logger.warning(
                            'Parameter "stop_at" is higher than video '
                            'duration, ignoring it.'
                        )
                        self.stop_at = -1
                except AttributeError:
                    logger.warning(
                        'Video duration could not be determined. Ignoring '
                        'parameters "start_at" and "stop_at"'
                    )
                    self.start_at = 0
                    self.stop_at = -1



from base64 import b64encode
from typing import List, Tuple

import cv2 as cv
import numpy as np


class Frame:
    def __init__(
        self,
        image: np.ndarray,
        key: int = None
    ):
        self.key: int = key
        self.data: dict = {}
        self.image: np.ndarray = image

    def serialize(self):
        return {
            'key': self.key,
            'image': self.image_bytes
        }

    @property
    def image_bytes(self):
        return cv.imencode('.jpg', self.image)[1]


class Face:
    def __init__(
        self,
        box: tuple,
        image: np.ndarray,
        aligned_image: np.ndarray = None,
        key: int = None,
        frame: Frame = None,
        subject: 'Subject' = None,
        landmarks: np.ndarray = None,
        embeddings: np.ndarray = None,
        detect_score: float = 0.0,
        mark_score: float = 0.0,
        nose_deviation: Tuple[float, float] = (0.0, 0.0),
        timestamp: float = 0,
        offset: Tuple[int, int] = (0, 0)
    ):
        self.key: int = key
        self.data: dict = {}
        self.box: tuple = box
        self.image: np.ndarray = image
        self.aligned_image: np.ndarray = aligned_image
        self.frame: Frame = frame
        self.subject: 'Subject' = subject
        self.embeddings: np.ndarray = embeddings
        self.landmarks: np.ndarray = landmarks
        self.detect_score: float = detect_score
        self.mark_score: float = mark_score
        self.nose_deviation: Tuple[float, float] = nose_deviation
        self.timestamp: float = timestamp
        self.offset: Tuple[int, int] = offset

    def serialize(self):

        frame = None
        if self.frame is not None:
            frame = self.frame.serialize()

        return {
            'box': self.box,
            'image': self.image_bytes,
            'frame': frame,
            'embeddings': self.embeddings.tolist(),
            'landmarks': self.landmarks.tolist(),
            'timestamp': self.timestamp
        }

    @property
    def image_bytes(self):
        return cv.imencode('.jpg', self.image)[1]

    @staticmethod
    def image_to_base64(image: np.ndarray) -> str:
        return str(b64encode(cv.imencode('.jpg', image)[1]), 'utf-8')

    def __str__(self):
        w, h = self.box[2] - self.box[0], self.box[3] - self.box[1]
        ret = f'score: {self.detect_score:.2f}, size: {w}x{h}, nose deviation: ' \
              f'({(self.nose_deviation[0] * 100):.2f} %, {(self.nose_deviation[1] * 100):.2f} %)'
        return ret

    def __repr__(self):
        return self.__str__()


class Subject:
    def __init__(
        self,
        faces: List[Face],
        embeddings: np.ndarray,
        key: int = None
    ):
        self.key: int = key
        self.data: dict = {}
        self.faces: List[Face] = faces
        self.embeddings: np.ndarray = embeddings
        self.last_updated: float = -float('inf')

    def append_face(self, face: Face):
        self.faces.append(face)
        if face.timestamp > self.last_updated:
            self.last_updated = face.timestamp

    def serialize(self):
        return {
            'faces': [face.serialize() for face in self.faces],
            'embeddings': self.embeddings.tolist()
        }



import torch
import math
import torch.nn as nn
from .resnet import *
from .l2norm import L2Norm


class CSPNet(nn.Module):
    def __init__(self):
        super(CSPNet, self).__init__()

        resnet = resnet50(pretrained=True, receptive_keep=True)

        self.conv1 = resnet.conv1
        self.bn1 = resnet.bn1
        self.relu = resnet.relu
        self.maxpool = resnet.maxpool
        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4

        self.p3 = nn.ConvTranspose2d(512, 256, kernel_size=4, stride=2, padding=1)
        self.p4 = nn.ConvTranspose2d(1024, 256, kernel_size=4, stride=4, padding=0)
        self.p5 = nn.ConvTranspose2d(2048, 256, kernel_size=4, stride=4, padding=0)

        nn.init.xavier_normal_(self.p3.weight)
        nn.init.xavier_normal_(self.p4.weight)
        nn.init.xavier_normal_(self.p5.weight)
        nn.init.constant_(self.p3.bias, 0)
        nn.init.constant_(self.p4.bias, 0)
        nn.init.constant_(self.p5.bias, 0)

        self.p3_l2 = L2Norm(256, 10)
        self.p4_l2 = L2Norm(256, 10)
        self.p5_l2 = L2Norm(256, 10)

        self.feat = nn.Conv2d(768, 256, kernel_size=3, stride=1, padding=1, bias=False)
        self.feat_bn = nn.BatchNorm2d(256, momentum=0.01)
        self.feat_act = nn.ReLU(inplace=True)

        self.pos_conv = nn.Conv2d(256, 1, kernel_size=1)
        self.reg_conv = nn.Conv2d(256, 1, kernel_size=1)
        self.off_conv = nn.Conv2d(256, 2, kernel_size=1)

        nn.init.xavier_normal_(self.feat.weight)
        nn.init.xavier_normal_(self.pos_conv.weight)
        nn.init.xavier_normal_(self.reg_conv.weight)
        nn.init.xavier_normal_(self.off_conv.weight)
        nn.init.constant_(self.pos_conv.bias, -math.log(0.99/0.01))
        nn.init.constant_(self.reg_conv.bias, 0)
        nn.init.constant_(self.off_conv.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)

        x = self.layer2(x)
        p3 = self.p3(x)
        p3 = self.p3_l2(p3)

        x = self.layer3(x)
        p4 = self.p4(x)
        p4 = self.p4_l2(p4)

        x = self.layer4(x)
        p5 = self.p5(x)
        p5 = self.p5_l2(p5)

        cat = torch.cat([p3, p4, p5], dim=1)

        feat = self.feat(cat)
        feat = self.feat_bn(feat)
        feat = self.feat_act(feat)

        x_cls = self.pos_conv(feat)
        x_cls = torch.sigmoid(x_cls)
        x_reg = self.reg_conv(feat)
        x_off = self.off_conv(feat)

        return x_cls, x_reg, x_off



class CSPNet_mod(nn.Module):
    # This is Batchnorm fixed version of CSP
    # under construction !!!!!!!!!!!!!!!!!!!
    def __init__(self):
        super(CSPNet_mod, self).__init__()

        resnet = resnet50(pretrained=True, receptive_keep=True)

        self.conv1 = resnet.conv1
        self.bn1 = resnet.bn1
        self.relu = resnet.relu
        self.maxpool = resnet.maxpool
        self.layer1 = resnet.layer1
        self.layer2 = resnet.layer2
        self.layer3 = resnet.layer3
        self.layer4 = resnet.layer4

        self.p3 = nn.ConvTranspose2d(512, 256, kernel_size=4, stride=2, padding=1)
        self.p4 = nn.ConvTranspose2d(1024, 256, kernel_size=4, stride=4, padding=0)
        self.p5 = nn.ConvTranspose2d(2048, 256, kernel_size=4, stride=4, padding=0)

        nn.init.xavier_normal_(self.p3.weight)
        nn.init.xavier_normal_(self.p4.weight)
        nn.init.xavier_normal_(self.p5.weight)
        nn.init.constant_(self.p3.bias, 0)
        nn.init.constant_(self.p4.bias, 0)
        nn.init.constant_(self.p5.bias, 0)

        self.p3_l2 = L2Norm(256, 10)
        self.p4_l2 = L2Norm(256, 10)
        self.p5_l2 = L2Norm(256, 10)

        self.feat = nn.Conv2d(768, 256, kernel_size=3, stride=1, padding=1, bias=True)
        self.feat_act = nn.ReLU(inplace=True)

        self.pos_conv = nn.Conv2d(256, 1, kernel_size=1)
        self.reg_conv = nn.Conv2d(256, 1, kernel_size=1)
        self.off_conv = nn.Conv2d(256, 2, kernel_size=1)

        nn.init.xavier_normal_(self.feat.weight)
        nn.init.xavier_normal_(self.pos_conv.weight)
        nn.init.xavier_normal_(self.reg_conv.weight)
        nn.init.xavier_normal_(self.off_conv.weight)

        nn.init.constant_(self.feat.bias, 0)
        nn.init.constant_(self.reg_conv.bias, -math.log(0.99/0.01))
        nn.init.constant_(self.pos_conv.bias, 0)
        nn.init.constant_(self.off_conv.bias, 0)

        for p in self.conv1.parameters():
            p.requires_grad = False
        for p in self.bn1.parameters():
            p.requires_grad = False
        for p in self.layer1.parameters():
            p.requires_grad = False

        def set_bn_fix(m):
            classname = m.__class__.__name__
            if classname.find('BatchNorm') != -1:
                for p in m.parameters(): p.requires_grad = False

        self.layer2.apply(set_bn_fix)
        self.layer3.apply(set_bn_fix)
        self.layer4.apply(set_bn_fix)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)

        x = self.layer2(x)
        p3 = self.p3(x)
        p3 = self.p3_l2(p3)

        x = self.layer3(x)
        p4 = self.p4(x)
        p4 = self.p4_l2(p4)

        x = self.layer4(x)
        p5 = self.p5(x)
        p5 = self.p5_l2(p5)

        cat = torch.cat([p3, p4, p5], dim=1)

        feat = self.feat(cat)
        feat = self.feat_act(feat)

        x_cls = self.pos_conv(feat)
        x_cls = torch.sigmoid(x_cls)
        x_reg = self.reg_conv(feat)
        x_off = self.off_conv(feat)

        return x_cls, x_reg, x_off

    def train(self, mode=True):
        # Override train so that the training mode is set as we want
        nn.Module.train(self, mode)
        if mode:
            # Set fixed blocks to be in eval mode
            self.conv1.eval()
            self.bn1.eval()
            self.layer1.eval()

            def set_bn_eval(m):
                classname = m.__class__.__name__
                if classname.find('BatchNorm') != -1:
                    m.eval()

            self.layer2.apply(set_bn_eval)
            self.layer3.apply(set_bn_eval)
            self.layer4.apply(set_bn_eval)



import pyflann
import numpy as np
from time import time
from multiprocessing import Pool
from functools import partial


def build_index(dataset, n_neighbors):
    """ Takes a dataset, returns the "n" nearest neighbors. """

    # Initialize FLANN
    pyflann.set_distance_type(distance_type='euclidean')
    flann = pyflann.FLANN()
    params = flann.build_index(dataset, algorithm='kdtree', trees=4)
    nearest_neighbors, dists = flann.nn_index(
        dataset, n_neighbors, checks=params['checks']
    )

    return nearest_neighbors, dists


def create_neighbor_lookup(nearest_neighbors):
    """
    Key is the reference face, values are the neighbors.
    """
    nn_lookup = {}
    for i in range(nearest_neighbors.shape[0]):
        nn_lookup[i] = nearest_neighbors[i, :]

    return nn_lookup


def calculate_symmetric_dist_row(nearest_neighbors, nn_lookup, row_no):
    """ Computes the symmetric distances for one row in the matrix. """

    dist_row = np.zeros([1, nearest_neighbors.shape[1]])
    f1 = nn_lookup[row_no]

    for idx, neighbor in enumerate(f1[1:]):
        Oi = idx+1
        co_neighbor = True
        try:
            row = nn_lookup[neighbor]
            Oj = np.where(row == row_no)[0][0] + 1
        except IndexError:
            Oj = nearest_neighbors.shape[1]+1
            co_neighbor = False

        # dij
        f11 = set(f1[0:Oi])
        f21 = set(nn_lookup[neighbor])
        dij = len(f11.difference(f21))
        # dji
        f12 = set(f1)
        f22 = set(nn_lookup[neighbor][0:Oj])
        dji = len(f22.difference(f12))

        if not co_neighbor:
            dist_row[0, Oi] = 9999.0
        else:
            dist_row[0, Oi] = float(dij + dji)/min(Oi, Oj)

    return dist_row


def calculate_symmetric_dist(app_nearest_neighbors):
    """ Computes the symmetric distance matrix. """

    dist_calc_time = time()
    nn_lookup = create_neighbor_lookup(app_nearest_neighbors)
    d = np.zeros(app_nearest_neighbors.shape)
    p = Pool(processes=4)
    func = partial(calculate_symmetric_dist_row, app_nearest_neighbors, nn_lookup)
    results = p.map(func, range(app_nearest_neighbors.shape[0]))
    for row_no, row_val in enumerate(results):
        d[row_no, :] = row_val
    d_time = time()-dist_calc_time
    print("Distance calculation time : {}".format(d_time))
    return d


def aro_cluster(
    features: [list, np.ndarray],
    n_neighbors: int = 10,
    distance_thr: float = 1.0
):
    """Approximate rank-order clustering.

    Takes in the nearest neighbors matrix and outputs clusters - list of lists.
    """

    app_nearest_neighbors, dists = build_index(features, n_neighbors)
    distance_matrix = calculate_symmetric_dist(app_nearest_neighbors)

    # Clustering :
    clusters = []

    # Start with the first face :
    nodes = set(list(np.arange(0, distance_matrix.shape[0])))

    plausible_neighbors = create_plausible_neighbor_lookup(
        app_nearest_neighbors, distance_matrix, distance_thr
    )

    while nodes:
        # Get a node :
        n = nodes.pop()

        # This contains the set of connected nodes :
        group = {n}

        # Build a queue with this node in it :
        queue = [n]

        # Iterate over the queue :
        while queue:
            n = queue.pop(0)
            neighbors = plausible_neighbors[n]
            # Remove neighbors we've already visited :
            neighbors = nodes.intersection(neighbors)
            neighbors.difference_update(group)

            # Remove nodes from the global set :
            nodes.difference_update(neighbors)

            # Add the connected neighbors :
            group.update(neighbors)

            # Add the neighbors to the queue to visit them next :
            queue.extend(neighbors)
        # Add the group to the list of groups :
        clusters.append(list(group))

    return clusters


def create_plausible_neighbor_lookup(
    app_nearest_neighbors,
    distance_matrix,
    thresh
):
    """Creates plausible neighbor lookup.

    Create a dictionary where the keys are the row numbers(face numbers) and
    the values are the plausible neighbors.
    """

    n_vectors = app_nearest_neighbors.shape[0]
    plausible_neighbors = {}
    for i in range(n_vectors):
        mask = distance_matrix[i, :] <= thresh
        plausible_neighbors[i] = set(
            list(app_nearest_neighbors[i, np.where(mask)][0])
        )
    return plausible_neighbors




from typing import List

import torch
import numpy as np
import cv2 as cv

from ._model import ONet

TRANSFORM_SIZE = (48, 48)


class FaceMarker:
    """Face marker.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.
    """

    def __init__(self, weights_path: str, force_cpu: bool = False):
        self.onet: ONet = ONet()

        weights: dict = np.load(weights_path, allow_pickle=True)[()]
        for n, p in self.onet.named_parameters():
            p.data = torch.tensor(weights[n], dtype=torch.float)

        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)
            self.onet = self.onet.to(self.gpu)

        self.onet.eval()

    def mark(self, images: List[np.ndarray]):
        """Estimate face landmark locations.

        Parameters
        ----------
        images : list of array_like images of length = n_faces
            A list of face images to be marked.

        Returns
        -------
        marks : array_like of shape = [n_faces, 5, 2]
            The estimated face landmarks. Each entry is the landmark set for
            the corresponding input face. Each entry in a landmark set contains
            the estimated (x, y) locations of the face landmarks in the
            following order: (0) left eye center, (1) right eye center,
            (2) nose center, (3) mouth left corner and (4) mouth right corner.
        scores : array_like of shape = [n_faces]
            The marking score for each estimated landmark set. The score is a
            number between 0 and 1 that represents the confidence of the
            landmarks estimation.
        """

        n_images = len(images)
        stack_shape = (n_images, 3) + TRANSFORM_SIZE
        image_stack = np.zeros(stack_shape, dtype=np.float32)
        image_sizes = np.zeros((n_images, 2), dtype=np.float32)

        for i in range(n_images):
            image_stack[i, :, :, :] = _image_transform(images[i])
            image_sizes[i, :] = images[i].shape[-2::-1]

        image_stack = torch.from_numpy(image_stack)
        if self.gpu is not None:
            image_stack = image_stack.to(self.gpu)

        with torch.no_grad():
            marks, offsets, scores = self.onet(image_stack)

        if self.gpu is not None:
            marks = marks.cpu().data.numpy()
            scores = scores.cpu().data.numpy()
        else:
            marks = marks.data.numpy()
            scores = scores.data.numpy()

        marks[:, 0:5] = np.expand_dims(image_sizes[:, 0], 1) * marks[:, 0:5]
        marks[:, 5:10] = np.expand_dims(image_sizes[:, 1], 1) * marks[:, 5:10]

        scores = scores[:, 1].reshape((-1,))
        marks = marks.reshape(-1, 2, 5).transpose(0, 2, 1).astype(np.int32)

        return marks, scores


def _image_transform(image: np.ndarray):
    image = cv.resize(image, TRANSFORM_SIZE, cv.INTER_LINEAR)
    image = image[:, :, ::-1]
    image = np.float32(image)
    image = image.transpose((2, 0, 1))
    image = np.expand_dims(image, 0)
    image = (image - 127.5) * 0.0078125
    return image



import json
from typing import Tuple, List
from validator_collection import validators
from .engine import VideoAnalyzer
from . import loggers


class InvalidSetting(Exception):
    pass


class InvalidJsonSettings(Exception):
    pass


class InvalidYamlSettings(Exception):
    pass


VIDEO_MODES = (VideoAnalyzer.MODE_ALL, VideoAnalyzer.MODE_HUNT)

LOGGING_LEVELS = (
    loggers.LEVEL_DEBUG,
    loggers.LEVEL_INFO,
    loggers.LEVEL_WARNING,
    loggers.LEVEL_ERROR,
    loggers.LEVEL_CRITICAL
)


class Settings:
    """Settings for FacesVision instance.

    See also
    --------
    vision.FacesVision
    """

    # noinspection PyTypeChecker
    def __init__(self):
        self._force_cpu: bool = False
        self._detector_weights_path: str = ''
        self._marker_weights_path: str = ''
        self._encoder_weights_path: str = ''
        self._genderage_weights_path: str = ''
        self._detection_min_height: int = 64
        self._detection_min_score: float = 0.9
        self._marking_min_score: float = 0.95
        self._detection_nms_thresh: float = 0.6
        self._detection_face_padding: float = 0.2
        self._detection_only: bool = False
        self._align_max_deviation: Tuple[float, float] = None
        self._face_align_size: int = 112
        self._max_frame_size: int = 0
        self._similarity_thresh: float = 0.5
        self._video_capture_source: [str, int, None] = None
        self._video_real_time: bool = True
        self._video_capture_resize_to: int = 0
        self._video_detect_interval: float = 1
        self._video_mode: str = 'all'
        self._video_hunt_embeddings: list = None
        self._video_hunt_keys: List[int] = None
        self._video_start_at: float = 0
        self._video_stop_at: float = -1
        self._video_roi_adapt: bool = False
        self._store_face_frames: bool = False
        self._faces_time_memory: float = 30
        self._logging_level: str = 'info'
        self._logging_file: str = ''
        self._log_to_console: bool = False

        self._fields = [attr[1::] for attr in self.__dict__]

    @property
    def force_cpu(self):
        """Use CPU instead GPU even if GPU is available

        Returns
        -------
        force_cpu : bool, (default=False)
        """
        return self._force_cpu

    @force_cpu.setter
    def force_cpu(self, val: bool):
        self._force_cpu = bool(validators.numeric(val))

    @property
    def detector_weights_path(self):
        """Path to file containing the model weights of face detector

        Returns
        -------
        detector_weights_path : str
        """
        return self._detector_weights_path

    @detector_weights_path.setter
    def detector_weights_path(self, val: str):
        self._detector_weights_path = validators.string(val)

    @property
    def marker_weights_path(self):
        """Path to file containing the model weights of face marker

        Returns
        -------
        marker_weights_path : str
        """
        return self._marker_weights_path

    @marker_weights_path.setter
    def marker_weights_path(self, val: str):
        self._marker_weights_path = validators.string(val)

    @property
    def encoder_weights_path(self):
        """Path to file containing the model weights of face encoder

        Returns
        -------
        encoder_weights_path : str
        """
        return self._encoder_weights_path

    @encoder_weights_path.setter
    def encoder_weights_path(self, val: str):
        self._encoder_weights_path = validators.string(val)

    @property
    def genderage_weights_path(self):
        """Path to file containing the model weights of gender-age predictor

        Returns
        -------
        genderage_weights_path : str
        """
        return self._genderage_weights_path

    @genderage_weights_path.setter
    def genderage_weights_path(self, val: str):
        self._genderage_weights_path = validators.string(val)

    @property
    def detection_min_height(self):
        """Minimum height in pixels of face to be detected

        Setting the detection minimum height to a reasonable value can help to
        save considerable computing power. Sometimes it is not necessary to
        detect very small faces because it will not be possible to recognize
        them.

        Returns
        -------
        detection_min_height : positive int (default=64)
        """
        return self._detection_min_height

    @detection_min_height.setter
    def detection_min_height(self, val: (float, int)):
        self._detection_min_height = int(validators.numeric(val, minimum=1))

    @property
    def detection_min_score(self):
        """Minimum confidence score of detected faces

        Faces with scores lower than the `detection_min_score` will be
        discarded.

        Returns
        -------
        detection_min_score : float in range [0, 1], (default=0.9)
        """
        return self._detection_min_score

    @detection_min_score.setter
    def detection_min_score(self, val: float):
        self._detection_min_score = validators.float(val, minimum=0, maximum=1)

    @property
    def marking_min_score(self):
        """Minimum marking score of a face after landmark detection

        Face analysis includes three steps: detection, alignment and
        encoding. At the alignment step, face landmarks are computed and a
        marking score between 0 and 1 is assigned to each face. Lower scores
        mean lower face alignment confidence. Faces with marking score lower
        than the `marking_min_score` will be discarded.

        Returns
        -------
        marking_min_score : float in range [0, 1], (default=0.95)
        """
        return self._marking_min_score

    @marking_min_score.setter
    def marking_min_score(self, val: float):
        self._marking_min_score = validators.float(val, minimum=0, maximum=1)

    @property
    def detection_nms_thresh(self):
        """Non-maximum suppression threshold of detected faces

        At detection step, from the set of face boxes which overlaps more than
        the corresponding NMS threshold the one with the highest score will be
        selected and the others discarded.

        Returns
        -------
        detection_nms_thresh : float in range [0, 1], (default=0.7)
        """
        return self._detection_nms_thresh

    @detection_nms_thresh.setter
    def detection_nms_thresh(self, val: float):
        self._detection_nms_thresh = validators.float(val, minimum=0, maximum=1)

    @property
    def detection_face_padding(self):
        """Relative padding of detected faces

        After a face is detected, the corresponding face image is extracted
        from the original image and sent to a face alignment stage. Padding the
        detected face box can help to obtain better alignment results. Padding
        value is a number between 0 and 1, relative to the face box height.

        Returns
        -------
        detection_face_padding : float in range [0, 1], (default=0.2)
        """
        return self._detection_face_padding

    @detection_face_padding.setter
    def detection_face_padding(self, val: float):
        self._detection_face_padding = validators.float(val, minimum=0, maximum=1)

    @property
    def detection_only(self):
        """Only detect faces, do no perform alignment or encoding

        Face analysis includes three steps: detection, alignment and
        encoding. If `detection_only = True`, only the detection step is done.

        Returns
        -------
        detection_only : bool, (default=False)
        """
        return self._detection_only

    @detection_only.setter
    def detection_only(self, val: bool):
        self._detection_only = bool(validators.numeric(val))

    @property
    def align_max_deviation(self):
        """Maximum deviation of the nose in a detected face

        Face analysis includes three steps: detection, alignment and
        encoding. At the alignment step, the nose deviation is computed. The
        nose deviation is a tuple where the first and second elements
        represents the horizontal and vertical deviation of nose landmark,
        respectively. If this deviation is higher than `align_max_deviation`
        (element-wise comparision), the face is discarded. For a deviation
        lower than (0.4, 0.3), the face is approximately frontal in both
        planes. Set it to `None` to allow any deviation.

        Returns
        -------
        align_max_deviation : tuple of length=2 or None, (default=None)
        """
        return self._align_max_deviation

    @align_max_deviation.setter
    def align_max_deviation(self, val: Tuple[float, float]):
        val = validators.iterable(
            val, minimum_length=2, maximum_length=2, allow_empty=True
        )
        self._align_max_deviation = val if val is None else tuple([
            validators.numeric(i, minimum=0) for i in val
        ])

    @property
    def face_align_size(self):
        """Output image size from face alignment

        Face analysis includes three steps: detection, alignment and
        encoding. The image out of the alignment step is squared with size
        `face_align_size`.

        Returns
        -------
        face_align_size : int, (default=112)
        """
        return self._face_align_size

    @face_align_size.setter
    def face_align_size(self, val: (float, int)):
        self._face_align_size = int(validators.numeric(val))

    @property
    def max_frame_size(self):
        """Maximum frame size

        During video analysis, when a face is detected a Face object is created.
        Optionally, the frame where the face was detected can be stored in the
        Face object. If a frame is stored, it is resized so that is maximum
        dimension is lowest than `max_frame_size`. Set it to a value lower than
        or equal to zero to do not resize.

        Returns
        -------
        max_frame_size : int, (default=0)
        """
        return self._max_frame_size

    @max_frame_size.setter
    def max_frame_size(self, val: (float, int)):
        self._max_frame_size = int(validators.numeric(val))

    @property
    def similarity_thresh(self):
        """Similarity threshold for face matching

        During face matching, two faces are considered to belong to the same
        person if the similarity between them is higher than
        `similarity_thresh`. A similarity of `1` means that the two faces are
        exactly the same, while a similarity of `0` means that the two faces
        are very different. Similarity between two faces if computed on the
        basis of the distance between their embeddings vectors.

        Returns
        -------
        similarity_thresh : float in range [0, 1], (default=0.5)
        """
        return self._similarity_thresh

    @similarity_thresh.setter
    def similarity_thresh(self, val: (float, int)):
        self._similarity_thresh = validators.numeric(val, minimum=0, maximum=1)

    ############################################################################
    # VIDEO ANALYSIS SETTINGS
    ############################################################################

    @property
    def video_capture_source(self):
        """Video analysis source

        For a string value, it must be a valid path to a video file of any
        supported format by `cv2.VideoCapture` class of opencv or a valid url
        video stream. For an int value, it refers to a connected camera,
        enumerated starting from zero. Set it to `None` to do initialize video
        processing engine.

        Returns
        -------
        video_capture_source : int, str or None (default=None)
        """
        return self._video_capture_source

    @video_capture_source.setter
    def video_capture_source(self, val: (str, int, None)):
        try:
            self._video_capture_source = validators.integer(
                val, coerce_value=True, allow_empty=True
            )
        except TypeError:
            self._video_capture_source = validators.string(
                val, allow_empty=True
            )

    @property
    def video_real_time(self):
        """Whether the video capture is from a camera or not

        Video analysis can be done from a real time source, like a camera, or
        from a non-real time source, like a video file. Although the video
        analysis for both types of sources is similar, some differences exists.
        For example, for a real time source when the algorithm finish analyzing
        one frame, it waits (sleeps) a time equal to `video_detect_interval`
        before processing a new frame. For a video file, the algorithms jumps
        directly into the next target frame after analyzing the current frame.

        Returns
        -------
        video_real_time : bool, (default=True)
        """
        return self._video_real_time

    @video_real_time.setter
    def video_real_time(self, val: bool):
        self._video_real_time = bool(validators.numeric(val))

    @property
    def video_capture_resize_to(self):
        """Size in pixels to which video frames are resized before processing

        After capturing the video frames, they are scaled so that its maximum
        dimension are equal to `video_capture_resize_to`. Set it to a value
        lower than or equal to zero to do not perform any resizing.

        Returns
        -------
        video_capture_resize_to : int, (default=0)
        """
        return self._video_capture_resize_to

    @video_capture_resize_to.setter
    def video_capture_resize_to(self, val: int):
        self._video_capture_resize_to = int(validators.numeric(val))

    @property
    def video_detect_interval(self):
        """Interval in seconds to analyze frames in video processing

        In order to save processing time, not all frames are analyzed during
        video processing. For a real time source, when the algorithm finish
        analyzing one frame, it waits (sleeps) a time equal to
        `video_detect_interval` before processing a new frame. Similarly, for a
        video file, the algorithms jumps directly into the next target frame
        located `video_detect_interval` seconds beyond. Setting it to a
        reasonable value can help to save considerable processing time, taking
        into account the (typical) low movement speed of people.

        Returns
        -------
        video_detect_interval : positive float, (default=1)
        """
        return self._video_detect_interval

    @video_detect_interval.setter
    def video_detect_interval(self, val: (int, float)):
        self._video_detect_interval = validators.numeric(val, minimum=0)

    @property
    def video_face_memory(self):
        """Time memory in seconds of faces during video analysis

        During video analysis, the algorithm try to match detected faces and
        group matched faces into Subject objects. A face is added to a Subject
        object only if no face have been added to that subject in the previous
        `face_memory` seconds. This setting helps to control the growth without
        limits of the number of faces added to a subject. Also, as face images
        belonging to the same person in successive nearby frames may be very
        similar, it also avoid storing redundant information.

        Returns
        -------
        video_face_memory : positive float, (default=30)
        """
        return self._faces_time_memory

    @video_face_memory.setter
    def video_face_memory(self, val: (float, int)):
        self._faces_time_memory = validators.numeric(val, minimum=0)

    @property
    def video_mode(self):
        """Video analysis mode

        Video analysis can be executed in two different modes: "all" or "hunt".
        In the "all" mode, the algorithm tries to register all the detected
        faces, grouping them in Subject objects. In the "hunt" mode the
        algorithm tries to match the detected faces with a predefined list of
        face embeddings and to register only matched faces.

        Note that if `video_mode` is set to "hunt", parameter `detection_only`
        will be ignored if set to `True`, because face encoding is needed to
        match against hunted faces.

        Returns
        -------
        video_mode : string, (default="all")
        """
        return self._video_mode

    @video_mode.setter
    def video_mode(self, val: str):
        val = validators.string(val)
        if val not in VIDEO_MODES:
            raise ValueError(f'"{val}" is not one of {VIDEO_MODES}.')
        self._video_mode = val

    @property
    def video_hunt_embeddings(self):
        """Embeddings vectors to hunt for

        If `video_mode = "hunt"` the algorithm tries to match the detected
        faces against a set of embeddings vectors in `hunt_embeddings` that
        represents the hunted faces, and to register only matched faces. Each
        embedding vector is of length 512, as returned by encoding.FaceEncoder.

        Returns
        -------
        video_hunt_embeddings : list or None, (default=None)
        """
        return self._video_hunt_embeddings

    @video_hunt_embeddings.setter
    def video_hunt_embeddings(self, val: List[List[float]]):
        self._video_hunt_embeddings = validators.iterable(
            val, allow_empty=True
        )

    @property
    def video_hunt_keys(self):
        """List of face keys associated to the hunt embeddings

        If `video_mode = "hunt"` the algorithm tries to match the detected
        faces against a set of embeddings vectors in `hunt_embeddings` that
        represents the hunted faces, and to register only matched faces. Each
        embedding vector of a hunted face must be associated to face key in
        `video_hunt_keys`, which is a list of integers numbers of length
        `n_hunts`.

        Returns
        -------
        video_hunt_keys : list of int or None, (default=None)
        """
        return self._video_hunt_keys

    @video_hunt_keys.setter
    def video_hunt_keys(self, val: list):
        self._video_hunt_keys = validators.iterable(val, allow_empty=True)

    @property
    def video_start_at(self):
        """Starting time in seconds of processing for video source files

        If `video_real_time = False`, the video file will be analyzed starting
        from `video_start_at` seconds. Note that if the duration of the video
        file is lower than `video_start_at` or `video_start_at` is higher than
        `video_stop_at`, `video_start_at` will be ignored.

        Returns
        -------
        video_start_at : float (default=0)
        """
        return self._video_start_at

    @video_start_at.setter
    def video_start_at(self, val: (float, int)):
        self._video_start_at = validators.numeric(val)

    @property
    def video_stop_at(self):
        """Stopping time in seconds of processing for video source files

        If `video_real_time = False`, the video file will analysis will be
        terminated at `video_stop_at` seconds. Note that if the duration of the
        video file is lower than `video_stop_at` or `video_start_at` is higher
        than `video_stop_at`, `video_stop_at` will be ignored. Set it to a
        negative value to run the analysis until the end of the video file.

        Returns
        -------
        video_stop_at : float (default=-1)
        """
        return self._video_stop_at

    @video_stop_at.setter
    def video_stop_at(self, val: (float, int)):
        self._video_stop_at = validators.numeric(val)

    @property
    def video_roi_adapt(self):
        """Use adaptive ROI during video analysis.

        If set to `True`, adaptive Region-Of-Interest (ROI) will be used during
        video analysis. This can help to save considerable processing time if
        the video comes from a static camera, as in these case the faces in the
        video frames are likely to be located inside a rectangle that changes
        little (in size and  location) over time. So, instead analyzing the
        entire frame, only a sub-region of the frame is analyzed. This
        rectangle is estimated at run time an used once it converges.

        Returns
        -------
        video_roi_adapt : bool (default=False)

        See also
        --------
        engine.AdaptiveRoi
        """
        return self._video_roi_adapt

    @video_roi_adapt.setter
    def video_roi_adapt(self, val: bool):
        self._video_roi_adapt = bool(validators.numeric(val))

    @property
    def store_face_frames(self):
        """Whether to store frames in faces or not

        During video analysis, when a face is detected a Face object is created.
        Optionally, the frame where the face was detected can be stored in the
        Face object if `store_face_frames` is set to `True`.

        Returns
        -------
        store_face_frames : bool (default=False)
        """
        return self._store_face_frames

    @store_face_frames.setter
    def store_face_frames(self, val: bool):
        self._store_face_frames = bool(validators.numeric(val))

    ############################################################################
    # LOGGING SETTINGS
    ############################################################################

    @property
    def logging_level(self):
        """Logging level

        Set the global logging level. Supported levels are "debug", "info",
        "warning", "error" or "critical".

        Returns
        -------
        logging_level : str (default="info")

        See also
        --------
        logging
        """
        return self._logging_level

    @logging_level.setter
    def logging_level(self, val: str):
        val = validators.string(val).lower()
        if val not in LOGGING_LEVELS:
            raise ValueError(f'"{val}" is not one of {LOGGING_LEVELS}.')
        self._logging_level = val

    @property
    def logging_file(self):
        """Path to logging file

        Path to a file to write logs. Set it to empty string to do not output
        logs to a file.

        Returns
        -------
        logging_file : str (default="")

        See also
        --------
        logging
        """
        return self._logging_file

    @logging_file.setter
    def logging_file(self, val: str):
        self._logging_file = validators.string(val)

    @property
    def log_to_console(self):
        """Whether to write logs to console or not

        If set to `True` all logs are written to the standard output.

        Returns
        -------
        logging_level : bool (default=False)

        See also
        --------
        logging
        """
        return self._log_to_console

    @log_to_console.setter
    def log_to_console(self, val: bool):
        self._log_to_console = bool(validators.numeric(val))

    ############################################################################
    # METHODS
    ############################################################################

    def serialize(self):
        data = {}
        for field in self._fields:
            data[field] = getattr(self, field)
        return data

    def dict_load(self, data: dict):
        data_fields = data.keys()
        for field in self._fields:
            if field in data_fields:
                setattr(self, field, data[field])

    def json_load(self, json_text):
        try:
            data = json.loads(json_text)
        except json.JSONDecodeError as err:
            raise InvalidJsonSettings(f'Can not decode json settings. {str(err)}.')

        if not isinstance(data, dict):
            raise InvalidJsonSettings(f'Decoded json settings is not a dictionary object.')

        self.dict_load(data)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models

AGE_CLS_COUNT = 60


class GenderAgeModel(torch.nn.Module):

    def __init__(self):
        super(GenderAgeModel, self).__init__()

        self.res_net = models.resnet18(pretrained=True)

        self.age_cls_unit = AGE_CLS_COUNT
        self.fc1 = nn.Linear(512, 512)
        self.age_cls_pred = nn.Linear(512, self.age_cls_unit)
        self.fc2 = nn.Linear(512, 512)
        self.gen_cls_pred = nn.Linear(512, 2)

    def get_resnet_convs_out(self, x):
        """Get outputs from convolutional layers of ResNet.

        Parameters
        ----------
        x : tensor-like
            Input image tensor.

        Returns
        -------
        x : Middle output from layer2, and final ouput from layer4.
        """

        x = self.res_net.conv1(x)  # out = [N, 64, 112, 112]
        x = self.res_net.bn1(x)
        x = self.res_net.relu(x)
        x = self.res_net.maxpool(x)  # out = [N, 64, 56, 56]

        x = self.res_net.layer1(x)  # out = [N, 64, 56, 56]
        x = self.res_net.layer2(x)  # out = [N, 128, 28, 28]
        x = self.res_net.layer3(x)  # out = [N, 256, 14, 14]
        x = self.res_net.layer4(x)  # out = [N, 512, 7, 7]

        return x  # out = [N, 512, 1 ,1]

    def get_age_gender(self, last_conv_out):

        last_conv_out = self.res_net.avgpool(last_conv_out)
        last_conv_out = last_conv_out.view(last_conv_out.size(0), -1)

        age_pred = F.relu(self.fc1(last_conv_out))
        age_pred = F.softmax(self.age_cls_pred(age_pred), 1)

        gen_pred = F.relu(self.fc2(last_conv_out))
        gen_pred = self.gen_cls_pred(gen_pred)

        return gen_pred, age_pred

    def forward(self, x):

        last1 = self.get_resnet_convs_out(x)
        gen_pred, age_pred = self.get_age_gender(last1)

        return gen_pred, age_pred



from typing import List

import boto3
import numpy as np
import cv2 as cv

MIN_IMAGE_WIDTH = 80
MAX_IMAGE_FACES = 100
MIN_FACE_WIDTH = 64
MAX_FACE_WIDTH = 164


class AwsFace:

    GENDER_MAN = 'male'
    GENDER_WOMAN = 'woman'

    def __init__(self):
        self.box: tuple = ()
        self.age_low: int = 0
        self.age_high: int = 0
        self.has_beard: bool = False
        self.beard_score: float = 0
        self.gender: str = ''
        self.gender_score: float = 0
        self.has_mustache: bool = False
        self.mustache_score: float = 0


_GENDER_CHOICES = {
    'Male': AwsFace.GENDER_MAN,
    'Female': AwsFace.GENDER_WOMAN
}


def _resize_image(image, min_width: int = None, max_width: int = None):
    height, width = image.shape[0:2]
    if max_width is not None and width > max_width:
        scale = max_width / width
        image = cv.resize(image, (max_width, int(scale * height)))
    elif min_width is not None and width < min_width:
        scale = min_width / width
        image = cv.resize(image, (min_width, int(scale * height)))

    return image


def _build_mosaic(images: List[np.ndarray]):

    n_images = len(images)
    images_ = []
    max_width = 0
    max_height = 0

    for image in images:
        image = _resize_image(
            image,
            min_width=MIN_FACE_WIDTH,
            max_width=MAX_FACE_WIDTH
        )
        images_.append(image)

        height, width = image.shape[0:2]
        if height > max_height:
            max_height = height
        if width > max_width:
            max_width = width

    images = images_

    n_rows = n_cols = int(n_images ** 0.5 + 1)
    if n_rows * (n_cols - 1) >= n_images:
        n_cols = n_cols - 1

    canvas_shape = (n_rows * max_height, n_cols * max_width, 3)
    image_mosaic = np.zeros(canvas_shape, dtype=np.uint8)

    image_ind = 0
    for row in range(n_rows):
        i = row * max_height
        for col in range(n_cols):
            j = col * max_width
            image = images[image_ind]
            height, width = image.shape[0:2]
            image_mosaic[i:(i + height), j:(j + width), :] = image[:]
            image_ind += 1

            if image_ind > n_images - 1:
                break

        if image_ind > n_images - 1:
            break

    return image_mosaic, n_rows, n_cols


def analyze_faces(images: List[np.ndarray]) -> List[AwsFace]:

    if not len(images):
        return []

    n_images = len(images)
    if n_images > 1:
        image_mosaic, n_rows, n_cols = _build_mosaic(images)
        image_mosaic = _resize_image(image_mosaic, min_width=MIN_IMAGE_WIDTH)
    else:
        image_mosaic = _resize_image(
            images[0],
            min_width=MIN_IMAGE_WIDTH,
            max_width=MAX_FACE_WIDTH
        )
        n_rows = n_cols = 1

    client = boto3.client('rekognition')

    _, image_bytes = cv.imencode('.jpg', image_mosaic)
    image_bytes = image_bytes.tobytes()

    response = client.detect_faces(
        Image={'Bytes': image_bytes},
        Attributes=['ALL']
    )

    faces = [None] * n_images

    mosaic_height, mosaic_width = image_mosaic.shape[:2]
    cell_width = mosaic_width / n_cols
    cell_height = mosaic_height / n_rows

    for df in response['FaceDetails']:
        box_top = float(df['BoundingBox']['Top'])
        box_left = float(df['BoundingBox']['Left'])
        box_width = float(df['BoundingBox']['Width'])
        box_height = float(df['BoundingBox']['Height'])

        row = int(mosaic_height * box_top / cell_height)
        col = int(mosaic_width * box_left / cell_width)
        image_ind = row * n_cols + col

        # cv.putText(
        #     image_mosaic,
        #     f'{row}, {col}, {image_ind}',
        #     (int(mosaic_width * box_left), int(mosaic_height * box_top)),
        #     color=(255,),
        #     fontFace=cv.FONT_HERSHEY_PLAIN,
        #     fontScale=1
        # )
        #
        # cv.rectangle(
        #     image_mosaic,
        #     pt1=(int(mosaic_width * box_left), int(mosaic_height * box_top)),
        #     pt2=(int(mosaic_width * (box_left + box_width)), int(mosaic_height * (box_top + box_height))),
        #     color=(255,),
        # )

        if image_ind < n_images:
            store = True
            if faces[image_ind] is not None:
                box = faces[image_ind].box
                area = (box[2] - box[0]) * (box[3] - box[1])
                if box_width * box_height < area:
                    store = False

            if store:
                face = AwsFace()
                face.box = (
                    box_left,
                    box_top,
                    box_left + box_width,
                    box_top + box_height,
                )
                face.age_low = int(df['AgeRange']['Low'])
                face.age_high = int(df['AgeRange']['High'])
                face.gender = _GENDER_CHOICES[str(df['Gender']['Value'])]
                face.gender_score = float(df['Gender']['Confidence'])/100
                face.has_beard = bool(df['Beard']['Value'])
                face.beard_score = float(df['Beard']['Confidence'])/100
                face.has_mustache = bool(df['Mustache']['Value'])
                face.mustache_score = float(df['Mustache']['Confidence'])/100
                # noinspection PyTypeChecker
                faces[image_ind] = face

            # cv.imshow('face', images[image_ind])
            # cv.waitKey()

    # cv.imshow('image', image_mosaic)
    # cv.waitKey()

    # noinspection PyTypeChecker
    return faces



import networkx as nx
from sklearn.neighbors import NearestNeighbors


def gcg_cluster(
    features,
    timestamps=None,
    distance_thr: float = 0.5,
    timestamp_thr: float = 0,
    edge_thr: float = 0.5
):
    n_features = len(features)
    dist_neigh = NearestNeighbors(radius=distance_thr)
    dist_neigh.fit(features)
    dist_graph = dist_neigh.radius_neighbors_graph(mode='distance')

    if timestamps is not None and timestamp_thr > 0:
        time_neigh = NearestNeighbors(radius=timestamp_thr)
        time_neigh.fit(timestamps)
        time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
        dist_graph = dist_graph.multiply(time_graph)
        dist_graph.eliminate_zeros()

    dist_graph = nx.from_scipy_sparse_matrix(dist_graph)

    # Dict of nodes that have not a cluster label assigned
    no_labels = {u: u for u in range(n_features)}

    # Init the clusters list with a random one-element cluster
    clusters = {0: [0]} # TODO: really do this random?

    # Remove node 0 form no labels list
    del no_labels[0]

    # Current growing cluster
    cur_label = 0

    node_boundary = nx.algorithms.boundary.node_boundary

    while True:
        cur_cluster = clusters[cur_label]
        boundary = node_boundary(dist_graph, cur_cluster, no_labels.keys())
        cluster_grow = False

        if len(boundary):
            scores = []
            for node in boundary:
                node_neighs = dist_graph[node]
                inside_edges = [
                    node_neighs[u]['weight'] for u in node_neighs
                    if u in cur_cluster
                ]
                scores.append((
                    len(inside_edges),
                    sum(inside_edges),
                    node
                ))

            best_score = sorted(scores, key=lambda s: (-s[0], s[1]))[0]
            if best_score[0] >= int(edge_thr * len(cur_cluster) + 1):
                best_node = best_score[2]
                clusters[cur_label].append(best_node)
                del no_labels[best_node]
                cluster_grow = True

        if not len(no_labels):
            break

        if not cluster_grow:
            # Increase current cluster label
            cur_label += 1
            # Take the next no labeled node and seed a new cluster with it
            next_node = no_labels.popitem()[0]
            clusters[cur_label] = [next_node]

            if not len(no_labels):
                break

    return list(clusters.values())


import torch.nn as nn
from torch.nn import init
from torchvision import models


def weights_init_kaiming(m):
    class_name = m.__class__.__name__
    if class_name.find('Conv') != -1:
        init.kaiming_normal_(m.weight.data, a=0, mode='fan_in')
    elif class_name.find('Linear') != -1:
        init.kaiming_normal_(m.weight.data, a=0, mode='fan_out')
        init.constant_(m.bias.data, 0.0)
    elif class_name.find('BatchNorm1d') != -1:
        init.normal_(m.weight.data, 1.0, 0.02)
        init.constant_(m.bias.data, 0.0)


def weights_init_classifier(m):
    class_name = m.__class__.__name__
    if class_name.find('Linear') != -1:
        init.normal_(m.weight.data, std=0.001)
        init.constant_(m.bias.data, 0.0)


class ClassBlock(nn.Module):
    """Defines the new fc layer and classification layer.

    |--Linear--|--bn--|--relu--|--Linear--|
    """

    def __init__(
        self,
        input_dim: int,
        class_num: int,
        droprate: float,
        relu: bool = False,
        bnorm: bool = True,
        num_bottleneck: int = 512,
        linear: bool = True,
        return_f: bool = False
    ):

        super(ClassBlock, self).__init__()
        self.return_f = return_f
        add_block = []
        if linear:
            add_block += [nn.Linear(input_dim, num_bottleneck)]
        else:
            num_bottleneck = input_dim
        if bnorm:
            add_block += [nn.BatchNorm1d(num_bottleneck)]
        if relu:
            add_block += [nn.LeakyReLU(0.1)]
        if droprate > 0:
            add_block += [nn.Dropout(p=droprate)]
        add_block = nn.Sequential(*add_block)
        add_block.apply(weights_init_kaiming)

        classifier = []
        classifier += [nn.Linear(num_bottleneck, class_num)]
        classifier = nn.Sequential(*classifier)
        classifier.apply(weights_init_classifier)

        self.add_block = add_block
        self.classifier = classifier

    def forward(self, x):
        x = self.add_block(x)
        if self.return_f:
            f = x
            x = self.classifier(x)
            return x, f
        else:
            x = self.classifier(x)
            return x


class DenseNet(nn.Module):
    """Define the DenseNet121-based Model.
    """

    def __init__(self, class_num, droprate=0.5):
        super().__init__()
        model_ft = models.densenet121(pretrained=True)
        model_ft.features.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        model_ft.fc = nn.Sequential()
        self.model = model_ft
        # For DenseNet, the feature dim is 1024
        self.classifier = ClassBlock(
            input_dim=1024,
            class_num=class_num,
            droprate=droprate
        )

    def forward(self, x):
        x = self.model.features(x)
        x = x.view(x.size(0), x.size(1))
        x = self.classifier(x)
        return x



import numpy as np
import torch
import cv2 as cv

from fnms import nms

from ._model import CSPNet

SCALE = np.float32([0.01712475, 0.017507, 0.01742919])
OFFSET = np.float32([2.11790393, 2.03571429, 1.80444444])


class PersonDetector:
    """Full-body person detector.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.
    """

    def __init__(
        self,
        weights_path: str,
        score_thresh: float = 0.2,
        nms_thresh: float = 0.3,
        resize_height: int = 256,
        force_cpu: bool = False
    ):
        self.score_thresh: float = score_thresh
        self.nms_thresh: float = nms_thresh
        self.resize_height: int = resize_height
        self.model: CSPNet = CSPNet()

        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)

        if self.gpu is None:
            cpu = torch.device('cpu')
            pretrained_dict = torch.load(weights_path, map_location=cpu)
        else:
            pretrained_dict = torch.load(weights_path)

        self.model.load_state_dict(pretrained_dict)

        if self.gpu is not None:
            self.model = self.model.to(self.gpu)

        self.model.eval()

    def detect(self, image: np.ndarray):
        """Predict gender and age of a list of face images.

        Parameters
        ----------
        image : array_like image
            A list of face images to be encoded.

        Returns
        -------
        genders : list of length = n_faces
            The predicted genders. Each entry represents the predicted gender
            of the corresponding entry in `images`.

        ages : list of length = n_faces
            The predicted ages. Each entry represents the predicted age
            of the corresponding entry in `images`.
        """

        image_h, image_w = image.shape[0:2]
        image, scale = _image_transform(image, self.resize_height, self.gpu)

        with torch.no_grad():
            # Eval model
            scores, heights, offsets = self.model(image)

        torch.cuda.synchronize(self.gpu)

        boxes, scores = _get_boxes(
            scores.cpu().numpy(),
            heights.cpu().numpy(),
            offsets.cpu().numpy(),
            (image_w, image_h),
            score_thresh=self.score_thresh,
            down=4,
            nms_thresh=self.nms_thresh
        )

        if scale != 1.0 and len(boxes):
            boxes = boxes / scale

        return np.int32(boxes), scores


def _image_transform(image: np.ndarray, resize_height: int, gpu=None):
    image_h, image_w = image.shape[0:2]
    scale = 1.0

    if image_h != resize_height:
        scale = resize_height / image_h
        image = cv.resize(image, (int(scale * image_w), resize_height))
        image_h, image_w = image.shape[0:2]

    if image_w % 32 != 0:
        pad_w = 32 - image_w % 32
        image = cv.copyMakeBorder(
            image, 0, 0, 0, pad_w,
            borderType=cv.BORDER_CONSTANT,
            value=(0, 0, 0)
        )

    image = image[:, :, ::-1]
    if image.dtype != np.float32:
        image = image.astype(np.float32)
    image = SCALE * image - OFFSET
    image = image.transpose((2, 0, 1))
    image = torch.from_numpy(image).unsqueeze(0)
    if gpu is not None:
        image = image.to(gpu)
    return image, scale


def _get_boxes(
    scores,
    heights,
    offsets,
    image_size,
    score_thresh=0.1,
    down=4,
    nms_thresh=0.3
):

    scores = np.squeeze(scores)
    heights = np.squeeze(heights)
    offset_y = offsets[0, 0, :, :]
    offset_x = offsets[0, 1, :, :]
    y_c, x_c = np.where(scores > score_thresh)
    boxes = []
    if len(y_c) > 0:
        for i in range(len(y_c)):
            h = np.exp(heights[y_c[i], x_c[i]]) * down
            w = 0.41 * h
            o_y = offset_y[y_c[i], x_c[i]]
            o_x = offset_x[y_c[i], x_c[i]]
            score = scores[y_c[i], x_c[i]]
            x1 = int(max(0, (x_c[i] + o_x + 0.5) * down - w / 2))
            y1 = int(max(0, (y_c[i] + o_y + 0.5) * down - h / 2))
            x2 = min(int(x1 + w), image_size[0] - 1)
            y2 = min(int(y1 + h), image_size[1] - 1)
            if x2 > x1 and y2 > y1:
                boxes.append((x1, y1, x2, y2, score))

        if len(boxes):
            boxes = np.float32(boxes)
            keep = nms(boxes, nms_thresh, force_cpu=True)
            boxes = boxes[keep, :]

    if len(boxes):
        return boxes[:, 0:4], boxes[:, 4]

    return [], []




from sklearn.neighbors import NearestNeighbors
from sklearn.cluster import DBSCAN, OPTICS
from scipy.sparse.csgraph import connected_components
from sklearn.cluster import AgglomerativeClustering
import networkx as nx


def cluster_labels(labels, indexes):
    labels_set = set(labels) - {-1}

    clusters = {label: [] for label in labels_set}
    outliers = []

    for index, label in zip(indexes, labels):
        if label != -1:
            clusters[label].append(index)
        else:
            outliers.append([index])

    return list(clusters.values()) + outliers


def hcg_cluster(
    features,
    timestamps=None,
    distance_thr: float = 0.5,
    timestamp_thr: float = 0,
    edge_thr: float = 0.7
):
    dist_neigh = NearestNeighbors(radius=distance_thr)
    dist_neigh.fit(features)
    dist_graph = dist_neigh.radius_neighbors_graph(mode='connectivity')

    if timestamps is not None and timestamp_thr > 0:
        time_neigh = NearestNeighbors(radius=timestamp_thr)
        time_neigh.fit(timestamps)
        time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
        dist_graph = dist_graph.multiply(time_graph)
        dist_graph.eliminate_zeros()

    dist_graph = nx.from_scipy_sparse_matrix(dist_graph)

    components = nx.connected_components(dist_graph)

    clusters = []

    clustering = None
    if distance_thr > 0:
        clustering = AgglomerativeClustering(
            n_clusters=None,
            distance_threshold=distance_thr,
            affinity='euclidean',
            linkage='ward'
        )

    for component in components:
        n = len(component)
        if n > 3 and clustering is not None:
            sub_graph = dist_graph.subgraph(component).copy()
            component = list(component)
            if sub_graph.size() >= edge_thr * (n * (n - 1) / 2):
                clusters.append(component)
            else:
                clustering.fit(features[component])
                clusters.extend(cluster_labels(clustering.labels_, component))
        else:
            clusters.append(list(component))

    return clusters


# def cluster_features(
#     features,
#     timestamps=None,
#     distance_thr: float = 0.5,
#     timestamp_thr: float = 0,
#     grouped: bool = True
# ):
#     dist_neigh = NearestNeighbors(radius=distance_thr)
#     dist_neigh.fit(features)
#     dist_graph = dist_neigh.radius_neighbors_graph(mode='connectivity')
#
#     if timestamps is not None and timestamp_thr > 0:
#         time_neigh = NearestNeighbors(radius=timestamp_thr)
#         time_neigh.fit(timestamps)
#         time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
#         dist_graph = dist_graph.multiply(time_graph)
#         dist_graph.eliminate_zeros()
#
#     clustering = AgglomerativeClustering(
#         n_clusters=None,
#         distance_threshold=distance_thr,
#         affinity='euclidean',
#         linkage='ward',
#         connectivity=dist_graph
#     )
#
#     clustering.fit(features)
#
#     labels = clustering.labels_
#
#     if not grouped:
#         return labels, None
#
#     labels_set = set(labels) - {-1}
#
#     clusters = {label: [] for label in labels_set}
#     outliers = []
#
#     for ind, label in enumerate(labels):
#         if label != -1:
#             clusters[label].append(ind)
#         else:
#             outliers.append([ind])
#
#     return labels, list(clusters.values()) + outliers

# import networkx as nx
# from .clustering import cluster_features as aro_cluster
#
# def cluster_features(
#     features,
#     timestamps=None,
#     distance_thr: float = 0.5,
#     timestamp_thr: float = 0,
#     grouped: bool = True
# ):
#     n_features = len(features)
#     dist_neigh = NearestNeighbors(radius=1.13)
#     dist_neigh.fit(features)
#     dist_graph = dist_neigh.radius_neighbors_graph(mode='connectivity')
#
#     if timestamps is not None and timestamp_thr > 0:
#         time_neigh = NearestNeighbors(radius=timestamp_thr)
#         time_neigh.fit(timestamps)
#         time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
#         dist_graph = dist_graph.multiply(time_graph)
#         dist_graph.eliminate_zeros()
#
#     dist_graph = nx.from_scipy_sparse_matrix(dist_graph)
#
#     components = nx.connected_components(dist_graph)
#
#     clusters = []
#
#     for component in components:
#         clusters.append(list(component))
#         n = len(component)
#         if n > 3:
#             sub_graph = dist_graph.subgraph(component).copy()
#             component = list(component)
#             if sub_graph.size() == (n * (n - 1) / 2):
#                 clusters.append(component)
#             else:
#                 sub_clusters = aro_cluster(
#                     features[component],
#                     distance_thr=distance_thr,
#                     n_neighbors=min(n, 10)
#                 )
#                 for sub_cluster in sub_clusters:
#                     clusters.append([component[i] for i in sub_cluster])
#         else:
#             clusters.append(list(component))
#
#     return clusters


# def cluster_features(
#     features,
#     timestamps=None,
#     distance_thr: float = 0.5,
#     timestamp_thr: float = 0,
#     grouped: bool = True
# ):
#     n_features = len(features)
#     dist_neigh = NearestNeighbors(radius=distance_thr)
#     dist_neigh.fit(features)
#     dist_graph = dist_neigh.radius_neighbors_graph(mode='connectivity')
#
#     if timestamps is not None and timestamp_thr > 0:
#         time_neigh = NearestNeighbors(radius=timestamp_thr)
#         time_neigh.fit(timestamps)
#         time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
#         dist_graph = dist_graph.multiply(time_graph)
#         dist_graph.eliminate_zeros()
#
#     _, labels = connected_components(
#         csgraph=dist_graph,
#         directed=False,
#         return_labels=True
#     )
#
#     clusters_high = cluster_labels(labels, range(0, n_features))
#
#     # Second level clustering
#     clusters_low = []
#     clustering = AgglomerativeClustering(
#         n_clusters=None,
#         distance_threshold=1.2,
#         affinity='euclidean',
#         linkage='single'
#     )
#
#     # clustering = OPTICS(
#     #     max_eps=distance_thr,
#     #     min_samples=3,
#     #     metric='euclidean'
#     # )
#
#     for cluster in clusters_high:
#         if len(cluster) > 3:
#             clustering.fit(features[cluster])
#             clusters_low.extend(
#                 cluster_labels(clustering.labels_, cluster)
#             )
#         else:
#             clusters_low.append(cluster)
#
#     return clusters_low


# import networkx as nx
#
# def cluster_features(
#     features,
#     timestamps=None,
#     distance_thr: float = 0.5,
#     timestamp_thr: float = 0,
#     grouped: bool = True
# ):
#     n_features = len(features)
#     dist_neigh = NearestNeighbors(radius=distance_thr)
#     dist_neigh.fit(features)
#     dist_graph = dist_neigh.radius_neighbors_graph(mode='connectivity')
#
#     if timestamps is not None and timestamp_thr > 0:
#         time_neigh = NearestNeighbors(radius=timestamp_thr)
#         time_neigh.fit(timestamps)
#         time_graph = time_neigh.radius_neighbors_graph(mode='connectivity')
#         dist_graph = dist_graph.multiply(time_graph)
#         dist_graph.eliminate_zeros()
#
#     dist_graph = nx.from_scipy_sparse_matrix(dist_graph)
#
#     components = nx.connected_components(dist_graph)
#
#     clusters = []
#
#     for component in components:
#         if len(component) > 3:
#             sub_graph = dist_graph.subgraph(component).copy()
#             sub_components = nx.k_edge_components(sub_graph, k=2)
#             clusters.extend([
#                 list(sub_component) for sub_component in sub_components
#             ])
#         else:
#             clusters.append(list(component))
#
#     return clusters


import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicConv2d(nn.Module):

    def __init__(self, in_channels, out_channels, **kwargs):
        super(BasicConv2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels, eps=1e-5)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        return F.relu(x, inplace=True)


class Inception(nn.Module):

    def __init__(self):
        super(Inception, self).__init__()
        self.branch1x1 = BasicConv2d(128, 32, kernel_size=1, padding=0)
        self.branch1x1_2 = BasicConv2d(128, 32, kernel_size=1, padding=0)
        self.branch3x3_reduce = BasicConv2d(128, 24, kernel_size=1, padding=0)
        self.branch3x3 = BasicConv2d(24, 32, kernel_size=3, padding=1)
        self.branch3x3_reduce_2 = BasicConv2d(128, 24, kernel_size=1, padding=0)
        self.branch3x3_2 = BasicConv2d(24, 32, kernel_size=3, padding=1)
        self.branch3x3_3 = BasicConv2d(32, 32, kernel_size=3, padding=1)

    def forward(self, x):
        branch1x1 = self.branch1x1(x)

        branch1x1_pool = F.avg_pool2d(x, kernel_size=3, stride=1, padding=1)
        branch1x1_2 = self.branch1x1_2(branch1x1_pool)

        branch3x3_reduce = self.branch3x3_reduce(x)
        branch3x3 = self.branch3x3(branch3x3_reduce)

        branch3x3_reduce_2 = self.branch3x3_reduce_2(x)
        branch3x3_2 = self.branch3x3_2(branch3x3_reduce_2)
        branch3x3_3 = self.branch3x3_3(branch3x3_2)

        outputs = [branch1x1, branch1x1_2, branch3x3, branch3x3_3]
        return torch.cat(outputs, 1)


class CRelu(nn.Module):

    def __init__(self, in_channels, out_channels, **kwargs):
        super(CRelu, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels, eps=1e-5)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = torch.cat([x, -x], 1)
        x = F.relu(x, inplace=True)
        return x


class FaceBoxes(nn.Module):

    def __init__(self, phase, size, num_classes):
        super(FaceBoxes, self).__init__()
        self.phase = phase
        self.num_classes = num_classes
        self.size = size

        self.conv1 = CRelu(3, 24, kernel_size=7, stride=4, padding=3)
        self.conv2 = CRelu(48, 64, kernel_size=5, stride=2, padding=2)

        self.inception1 = Inception()
        self.inception2 = Inception()
        self.inception3 = Inception()

        self.conv3_1 = BasicConv2d(128, 128, kernel_size=1, stride=1, padding=0)
        self.conv3_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)

        self.conv4_1 = BasicConv2d(256, 128, kernel_size=1, stride=1, padding=0)
        self.conv4_2 = BasicConv2d(128, 256, kernel_size=3, stride=2, padding=1)

        self.loc, self.conf = self.multibox(self.num_classes)

        if self.phase == 'test':
            self.softmax = nn.Softmax(dim=-1)

        if self.phase == 'train':
            for m in self.modules():
                if isinstance(m, nn.Conv2d):
                    if m.bias is not None:
                        nn.init.xavier_normal_(m.weight.data)
                        m.bias.data.fill_(0.02)
                    else:
                        m.weight.data.normal_(0, 0.01)
                elif isinstance(m, nn.BatchNorm2d):
                    m.weight.data.fill_(1)
                    m.bias.data.zero_()

    def multibox(self, num_classes):
        loc_layers = []
        conf_layers = []
        loc_layers += [nn.Conv2d(128, 21 * 4, kernel_size=3, padding=1)]
        conf_layers += [nn.Conv2d(128, 21 * num_classes, kernel_size=3, padding=1)]
        loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
        conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
        loc_layers += [nn.Conv2d(256, 1 * 4, kernel_size=3, padding=1)]
        conf_layers += [nn.Conv2d(256, 1 * num_classes, kernel_size=3, padding=1)]
        return nn.Sequential(*loc_layers), nn.Sequential(*conf_layers)

    def forward(self, x):

        detection_sources = list()
        loc = list()
        conf = list()

        x = self.conv1(x)
        x = F.max_pool2d(x, kernel_size=3, stride=2, padding=1)
        x = self.conv2(x)
        x = F.max_pool2d(x, kernel_size=3, stride=2, padding=1)
        x = self.inception1(x)
        x = self.inception2(x)
        x = self.inception3(x)
        detection_sources.append(x)

        x = self.conv3_1(x)
        x = self.conv3_2(x)
        detection_sources.append(x)

        x = self.conv4_1(x)
        x = self.conv4_2(x)
        detection_sources.append(x)

        for (x, l, c) in zip(detection_sources, self.loc, self.conf):
            loc.append(l(x).permute(0, 2, 3, 1).contiguous())
            conf.append(c(x).permute(0, 2, 3, 1).contiguous())

        loc = torch.cat([o.view(o.size(0), -1) for o in loc], 1)
        conf = torch.cat([o.view(o.size(0), -1) for o in conf], 1)

        if self.phase == "test":
            output = (loc.view(loc.size(0), -1, 4),
                      self.softmax(conf.view(conf.size(0), -1, self.num_classes)))
        else:
            output = (loc.view(loc.size(0), -1, 4),
                      conf.view(conf.size(0), -1, self.num_classes))

        return output



from typing import List

import cv2
import numpy as np
import torch

from ._model import GenderAgeModel, AGE_CLS_COUNT

RESIZE = 256
CROP = 224
SCALE = np.float32([0.01712475, 0.017507, 0.01742919])
OFFSET = np.float32([2.11790393, 2.03571429, 1.80444444])

RESIZE_SHAPE = (RESIZE, RESIZE)

CROP_RECT = (
    int((RESIZE - CROP)/2),
    int((RESIZE - CROP)/2),
    int((RESIZE + CROP)/2),
    int((RESIZE + CROP)/2),
)

AGE_CLS_WEIGHTS = np.linspace(1, AGE_CLS_COUNT, AGE_CLS_COUNT, dtype=np.float32)


class GenderAgePredictor:
    """Gender and age predictor.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.
    """

    GENDER_WOMAN = 0
    GENDER_MAN = 1

    def __init__(
        self,
        weights_path: str,
        force_cpu: bool = False
    ):
        self.model: GenderAgeModel = GenderAgeModel()

        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)

        if self.gpu is None:
            cpu = torch.device('cpu')
            pretrained_dict = torch.load(weights_path, map_location=cpu)
        else:
            pretrained_dict = torch.load(weights_path)

        self.model.load_state_dict(pretrained_dict)

        if self.gpu is not None:
            self.model = self.model.to(self.gpu)

        self.model.eval()

    def predict(self, images: List[np.ndarray]):
        """Predict gender and age of a list of face images.

        Parameters
        ----------
        images : list of array_like images of length = n_faces
            A list of face images to be encoded.

        Returns
        -------
        genders : list of length = n_faces
            The predicted genders. Each entry represents the predicted gender
            of the corresponding entry in `images`.

        ages : list of length = n_faces
            The predicted ages. Each entry represents the predicted age
            of the corresponding entry in `images`.
        """

        shape = (len(images), 3, CROP, CROP)
        batch = np.zeros(shape, dtype=np.float32)

        for index, image in enumerate(images):
            image = _image_transform(image)
            batch[index, :, :, :] = image

        batch_tensor = torch.from_numpy(batch)
        if self.gpu is not None:
            batch_tensor = batch_tensor.to(self.gpu)

        with torch.no_grad():
            # Eval model
            genders_out, ages_probs = self.model(batch_tensor)

            genders_out = torch.softmax(genders_out, dim=1)
            genders_probs, genders_preds = torch.max(genders_out, dim=1)
            genders_probs = genders_probs.cpu().data.numpy()
            genders_preds = genders_preds.cpu().data.numpy()

            ages_probs = ages_probs.cpu().data.numpy()
            ages_probs = ages_probs.reshape((-1, AGE_CLS_COUNT))
            ages_preds = np.sum(ages_probs * AGE_CLS_WEIGHTS, axis=1)

            diff = AGE_CLS_WEIGHTS - ages_preds.reshape((-1, 1))
            age_vars = np.sqrt(np.mean(ages_probs * diff * diff, axis=1))

        return genders_preds, genders_probs, ages_preds, age_vars


def _image_transform(image: np.ndarray):

    # Resize
    if image.shape[0:2] != RESIZE_SHAPE:
        image = cv2.resize(image, (RESIZE, RESIZE))
    # Crop
    image = image[CROP_RECT[1]:CROP_RECT[3], CROP_RECT[0]:CROP_RECT[2]]
    # BGR to RGB
    image = image[:, :, ::-1]
    # Convert to float
    if image.dtype != np.float32:
        image = np.float32(image)
    # Normalize
    image = SCALE * image - OFFSET # [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    # Set tensor shape
    image = image.transpose((2, 0, 1))
    image = np.reshape(image, (1, 3, CROP, CROP))

    return image




# Original work Copyright (c) 2019 Jian Zhao
# Copyright (c) 2019, Raikel Bordon.
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file.

import torch.nn as nn
import torch.nn.functional as nn_func
from collections import OrderedDict


class Flatten(nn.Module):

    def __init__(self):
        super(Flatten, self).__init__()

    def forward(self, x):
        """
        Arguments:
            x: a float tensor with shape [batch_size, c, h, w].
        Returns:
            a float tensor with shape [batch_size, c*h*w].
        """

        x = x.transpose(3, 2).contiguous()
        return x.view(x.size(0), -1)


class ONet(nn.Module):

    def __init__(self):

        super(ONet, self).__init__()

        self.features = nn.Sequential(OrderedDict([
            ('conv1', nn.Conv2d(3, 32, 3, 1)),
            ('prelu1', nn.PReLU(32)),
            ('pool1', nn.MaxPool2d(3, 2, ceil_mode=True)),

            ('conv2', nn.Conv2d(32, 64, 3, 1)),
            ('prelu2', nn.PReLU(64)),
            ('pool2', nn.MaxPool2d(3, 2, ceil_mode=True)),

            ('conv3', nn.Conv2d(64, 64, 3, 1)),
            ('prelu3', nn.PReLU(64)),
            ('pool3', nn.MaxPool2d(2, 2, ceil_mode=True)),

            ('conv4', nn.Conv2d(64, 128, 2, 1)),
            ('prelu4', nn.PReLU(128)),

            ('flatten', Flatten()),
            ('conv5', nn.Linear(1152, 256)),
            ('drop5', nn.Dropout(0.25)),
            ('prelu5', nn.PReLU(256)),
        ]))

        self.conv6_1 = nn.Linear(256, 2)
        self.conv6_2 = nn.Linear(256, 4)
        self.conv6_3 = nn.Linear(256, 10)

    def forward(self, x):
        """
        Arguments:
            x: a float tensor with shape [batch_size, 3, h, w].
        Returns:
            c: a float tensor with shape [batch_size, 10].
            b: a float tensor with shape [batch_size, 4].
            a: a float tensor with shape [batch_size, 2].
        """
        x = self.features(x)
        a = self.conv6_1(x)
        b = self.conv6_2(x)
        c = self.conv6_3(x)
        a = nn_func.softmax(a, dim=1)
        return c, b, a



import torch
import torch.nn as nn
import torch.nn.init as init


class L2Norm(nn.Module):
    def __init__(self, n_channels, scale):
        super(L2Norm, self).__init__()
        self.n_channels = n_channels
        self.gamma = scale or None
        self.eps = 1e-10
        self.weight = nn.Parameter(torch.Tensor(self.n_channels))
        self.reset_parameters()

    def reset_parameters(self):
        init.constant_(self.weight, self.gamma)

    def forward(self, x):
        norm = x.pow(2).sum(dim=1, keepdim=True).sqrt()+self.eps
        x = torch.div(x, norm)
        out = self.weight.unsqueeze(0).unsqueeze(2).unsqueeze(3).expand_as(x) * x
        return out



from typing import List

import cv2
import numpy as np
import torch

from ._model import IR_50

DEFAULT_INPUT_SIZE = 112
USE_TTA = True


class FaceEncoder:
    """Face encoder.

    Parameters
    ----------
    weights_path : str
        Absolute path to file containing pretrained model weights.
    """

    def __init__(
        self,
        weights_path: str,
        force_cpu: bool = False,
        input_size: int = DEFAULT_INPUT_SIZE
    ):
        self.input_size: int = input_size
        self.model: IR_50 = IR_50((input_size, input_size))

        # noinspection PyTypeChecker
        self.gpu: torch.device = None
        if torch.cuda.is_available() and not force_cpu:
            self.gpu = torch.device('cuda', 0)

        if self.gpu is None:
            cpu = torch.device('cpu')
            pretrained_dict = torch.load(weights_path, map_location=cpu)
        else:
            pretrained_dict = torch.load(weights_path)

        self.model.load_state_dict(pretrained_dict)

        if self.gpu is not None:
            self.model = self.model.to(self.gpu)

        self.model.eval()

    def encode(self, images: List[np.ndarray]):
        """Encode face images.

        Parameters
        ----------
        images : list of array_like images of length = n_faces
            A list of face images to be encoded.

        Returns
        -------
        embeddings : array_like of shape = [n_faces, 512]
            Embeddings vectors. Each entry is the embedding vector of the
            corresponding input face image.
        """

        image_shape = (self.input_size, self.input_size)
        if USE_TTA:
            shape = (2 * len(images), 3) + image_shape
            batch = np.zeros(shape, dtype=np.float32)
        else:
            shape = (len(images), 3) + image_shape
            batch = np.zeros(shape, dtype=np.float32)

        for index, image in enumerate(images):
            image = cv2.resize(image, image_shape)

            image_flipped = None
            if USE_TTA:
                image_flipped = _image_transform(
                    cv2.flip(image, 1), self.input_size
                )

            image = _image_transform(image, self.input_size)

            if USE_TTA:
                batch[2 * index, :, :, :] = image
                batch[2 * index + 1, :, :, :] = image_flipped
            else:
                batch[index, :, :, :] = image

        batch_tensor = torch.from_numpy(batch)
        if self.gpu is not None:
            batch_tensor = batch_tensor.to(self.gpu)

        with torch.no_grad():
            emb_batch = self.model(batch_tensor)
            if USE_TTA:
                emb_batch = emb_batch[0::2] + emb_batch[1::2]

            embeddings = _l2_norm(emb_batch)

        if self.gpu is not None:
            embeddings = embeddings.cpu()

        return embeddings.numpy()


def _image_transform(image: np.ndarray, size: int):
    # BGR to RGB
    image = image[:, :, ::-1]
    # load numpy to tensor
    image = image.transpose((2, 0, 1))
    image = np.reshape(image, (1, 3, size, size))
    if image.dtype != np.float32:
        image = np.float32(image)
    image = (image - 127.5) * 0.0078125
    return image


def _l2_norm(embeddings, axis=1):
    return torch.div(embeddings, torch.norm(embeddings, 2, axis, True))



# Original work Copyright (c) 2019 Jian Zhao
# Copyright (c) 2019, Raikel Bordon.
# Use of this source code is governed by a MIT-style license that can be
# found in the LICENSE file.

from collections import namedtuple

import torch
import torch.nn as nn
from torch.nn import (
    Linear,
    Conv2d,
    BatchNorm1d,
    BatchNorm2d,
    PReLU,
    ReLU,
    Sigmoid,
    Dropout,
    MaxPool2d,
    AdaptiveAvgPool2d,
    Sequential,
    Module
)


class Flatten(Module):
    def forward(self, input):
        return input.view(input.size(0), -1)


def l2_norm(input, axis=1):
    norm = torch.norm(input, 2, axis, True)
    output = torch.div(input, norm)

    return output


class SEModule(Module):
    def __init__(self, channels, reduction):
        super(SEModule, self).__init__()
        self.avg_pool = AdaptiveAvgPool2d(1)
        self.fc1 = Conv2d(
            channels, channels // reduction, kernel_size=1, padding=0, bias=False)

        nn.init.xavier_uniform_(self.fc1.weight.data)

        self.relu = ReLU(inplace=True)
        self.fc2 = Conv2d(
            channels // reduction, channels, kernel_size=1, padding=0, bias=False)

        self.sigmoid = Sigmoid()

    def forward(self, x):
        module_input = x
        x = self.avg_pool(x)
        x = self.fc1(x)
        x = self.relu(x)
        x = self.fc2(x)
        x = self.sigmoid(x)

        return module_input * x


class BottleneckIR(Module):
    def __init__(self, in_channel, depth, stride):
        super(BottleneckIR, self).__init__()
        if in_channel == depth:
            self.shortcut_layer = MaxPool2d(1, stride)
        else:
            self.shortcut_layer = Sequential(
                Conv2d(in_channel, depth, (1, 1), stride, bias=False), BatchNorm2d(depth))
        self.res_layer = Sequential(
            BatchNorm2d(in_channel),
            Conv2d(in_channel, depth, (3, 3), (1, 1), 1, bias=False), PReLU(depth),
            Conv2d(depth, depth, (3, 3), stride, 1, bias=False), BatchNorm2d(depth))

    def forward(self, x):
        shortcut = self.shortcut_layer(x)
        res = self.res_layer(x)

        return res + shortcut


class BottleneckIRSE(Module):
    def __init__(self, in_channel, depth, stride):
        super(BottleneckIRSE, self).__init__()
        if in_channel == depth:
            self.shortcut_layer = MaxPool2d(1, stride)
        else:
            self.shortcut_layer = Sequential(
                Conv2d(in_channel, depth, (1, 1), stride, bias=False),
                BatchNorm2d(depth))
        self.res_layer = Sequential(
            BatchNorm2d(in_channel),
            Conv2d(in_channel, depth, (3, 3), (1, 1), 1, bias=False),
            PReLU(depth),
            Conv2d(depth, depth, (3, 3), stride, 1, bias=False),
            BatchNorm2d(depth),
            SEModule(depth, 16)
        )

    def forward(self, x):
        shortcut = self.shortcut_layer(x)
        res = self.res_layer(x)

        return res + shortcut


class Bottleneck(namedtuple('Block', ['in_channel', 'depth', 'stride'])):
    """A named tuple describing a ResNet block."""


def get_block(in_channel, depth, num_units, stride=2):
    return [Bottleneck(in_channel, depth, stride)] + [
        Bottleneck(depth, depth, 1) for i in range(num_units - 1)
    ]


def get_blocks(num_layers):
    if num_layers == 50:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=4),
            get_block(in_channel=128, depth=256, num_units=14),
            get_block(in_channel=256, depth=512, num_units=3)
        ]
    elif num_layers == 100:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=13),
            get_block(in_channel=128, depth=256, num_units=30),
            get_block(in_channel=256, depth=512, num_units=3)
        ]
    elif num_layers == 152:
        blocks = [
            get_block(in_channel=64, depth=64, num_units=3),
            get_block(in_channel=64, depth=128, num_units=8),
            get_block(in_channel=128, depth=256, num_units=36),
            get_block(in_channel=256, depth=512, num_units=3)
        ]
    else:
        raise ValueError(f'Invalid number of layers: {num_layers}')

    return blocks


class Backbone(Module):
    def __init__(self, input_size, num_layers, mode='ir'):
        super(Backbone, self).__init__()
        assert input_size[0] in [112, 224], "input_size should be [112, 112] or [224, 224]"
        assert num_layers in [50, 100, 152], "num_layers should be 50, 100 or 152"
        assert mode in ['ir', 'ir_se'], "mode should be ir or ir_se"
        blocks = get_blocks(num_layers)
        if mode == 'ir':
            unit_module = BottleneckIR
        elif mode == 'ir_se':
            unit_module = BottleneckIRSE
        self.input_layer = Sequential(Conv2d(3, 64, (3, 3), 1, 1, bias=False),
                                      BatchNorm2d(64),
                                      PReLU(64))
        if input_size[0] == 112:
            self.output_layer = Sequential(BatchNorm2d(512),
                                           Dropout(),
                                           Flatten(),
                                           Linear(512 * 7 * 7, 512),
                                           BatchNorm1d(512))
        else:
            self.output_layer = Sequential(BatchNorm2d(512),
                                           Dropout(),
                                           Flatten(),
                                           Linear(512 * 14 * 14, 512),
                                           BatchNorm1d(512))

        modules = []
        for block in blocks:
            for bottleneck in block:
                modules.append(
                    unit_module(bottleneck.in_channel,
                                bottleneck.depth,
                                bottleneck.stride))
        self.body = Sequential(*modules)

        self._initialize_weights()

    def forward(self, x):
        x = self.input_layer(x)
        x = self.body(x)
        x = self.output_layer(x)

        return x

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.BatchNorm1d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight.data)
                if m.bias is not None:
                    m.bias.data.zero_()


def IR_50(input_size):
    """Constructs a ir-50 model.
    """
    model = Backbone(input_size, 50, 'ir')

    return model


def IR_101(input_size):
    """Constructs a ir-101 model.
    """
    model = Backbone(input_size, 100, 'ir')

    return model


def IR_152(input_size):
    """Constructs a ir-152 model.
    """
    model = Backbone(input_size, 152, 'ir')

    return model


def IR_SE_50(input_size):
    """Constructs a ir_se-50 model.
    """
    model = Backbone(input_size, 50, 'ir_se')

    return model


def IR_SE_101(input_size):
    """Constructs a ir_se-101 model.
    """
    model = Backbone(input_size, 100, 'ir_se')

    return model


def IR_SE_152(input_size):
    """Constructs a ir_se-152 model.
    """
    model = Backbone(input_size, 152, 'ir_se')

    return model



import cv2 as cv
import numpy as np

# Default reference facial points for crop size of 112x112
REFERENCE_POINTS = np.float32([
    [38.29459953, 51.69630051],
    [73.53179932, 51.50139999],
    [56.02519989, 71.73660278],
    [41.54930115, 92.3655014],
    [70.72990036, 92.20410156]
])

REFERENCE_SIZE = 112


class FaceAligner:
    """Face aligner.

    Parameters
    ----------

    out_size : int, optional, (default=112)
        Size of output aligned image.
    """

    def __init__(self, out_size: int = 112):
        scale = out_size / REFERENCE_SIZE
        self._reference: np.ndarray = scale * REFERENCE_POINTS
        self.out_size: int = out_size

    def align(self, image: np.ndarray, marks: np.ndarray):
        """Aligns a face image.

        Parameters
        ----------
        image : array_like
            The image face to be aligned.

        marks : array_like of shape = [5, 2]
            The landmark set of the input image face. Each entry in `marks`
            must contains the (x, y) locations of the face landmarks in the
            following order: (0) left eye center, (1) right eye center, (2)
            nose center, (3) mouth left corner and (4) mouth right corner.

        Returns
        -------
        aligned_image : array_like of shape = [out_size, out_size, 3]
            The resulting aligned face image. Alignment is made by computing
            the optimal limited affine transformation with 4 degrees of freedom
            between the set of input face landmarks and a set of reference
            points that represents a reference front-aligned face.
        nose_deviation : tuple of length = 2
            The nose deviation of the input face image. The nose deviation is a
            tuple where the first and second elements represents the horizontal
            and vertical deviation of nose landmark, respectively For a
            deviation lower than (0.4, 0.3), the face is approximately frontal
            in both planes.
        """

        if marks.dtype != np.float32:
            marks = np.float32(marks)

        tfm, _ = cv.estimateAffinePartial2D(
            marks,
            self._reference.copy(),
            method=cv.LMEDS
        )
        aligned_image = cv.warpAffine(
            image, tfm, (self.out_size, self.out_size)
        )
        aligned_marks = (
            np.matmul(tfm[:, 0:2], marks.T) +
            tfm[:, 2].reshape((-1, 1))
        ).T

        face_center = np.mean(aligned_marks[[0, 1, 3, 4]], axis=0)
        eye_dist = abs(aligned_marks[0][0] - aligned_marks[1][0])
        nose_x, nose_y = aligned_marks[2]
        nose_deviation = (
            abs(face_center[0] - nose_x) / eye_dist,
            abs(face_center[1] - nose_y) / eye_dist,
        )

        return aligned_image, nose_deviation
